from pathlib import Path
from typing import List

import pandas as pd

import dtiplayground.dmri.common as common
import dtiplayground.dmri.fiberprofile as base
from dtiplayground.dmri.common import tools

logger = common.logger.write


class EXTRACT_Profile(base.modules.DTIFiberProfileModule):
    def __init__(self, config_dir, *args, **kwargs):
        super().__init__(config_dir)

    def generateDefaultProtocol(self, image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    """
    Main function called by the pipeline when this module runs.
    """

    def process(self, *args, **kwargs):
        super().process()
        inputParams = self.getPreviousResult()['output']
        protocol_options = args[0]
        self.software_info = protocol_options['software_info']['softwares']

        # Reading some parameters
        path_to_csv: str = inputParams["file_path"]
        output_base_dir: str = self.output_dir  # output directory string
        atlas_path: str = self.protocol["atlas"]
        tracts: List[str] = self.protocol["tracts"].split(',')
        properties_to_profile: List[str] = [x.strip() for x in self.protocol["propertiesToProfile"].split(',')]
        result_case_columnwise: bool = self.protocol["resultCaseColumnwise"]
        input_is_dti: bool = self.protocol["inputIsDTI"]
        overwrite: bool = self.options['overwrite']
        use_displacement_field: bool = self.protocol["useDisplacementField"]
        step_size: str = str(self.protocol["stepSize"])
        plane_of_origin: str = self.protocol["planeOfOrigin"]
        support_bandwidth: str = str(self.protocol["supportBandwidth"])
        noNaN: str = self.protocol["noNaN"]
        mask: str = self.protocol["mask"]

        df = pd.read_csv(path_to_csv)

        recompute_scalars: bool = overwrite


        # Get parameter to col map, overriding with user inputs if necessary
        # Generate default map
        parameter_to_col_map = {'Case ID': 'id', 'Original DTI Image': 'Original DTI image',
                                'Deformation Field': 'Concatenated Deformation field'}
        for scalar in ['FA', 'MD', 'AD', 'RD']:
            scalar_col = f'{scalar} from original'
            parameter_to_col_map[scalar] = scalar_col
        for scalar in properties_to_profile:
            scalar_col = f'{scalar} from original'
            parameter_to_col_map[scalar] = scalar_col

        # Update defaults with user overrides
        user_parameter_to_col_map = self.protocol['parameterToColumnHeaderMap']
        if user_parameter_to_col_map is not None:
            parameter_to_col_map.update(user_parameter_to_col_map)

        if input_is_dti:
            # check to see if the scalar images have already been generated
            # if not, generate them

            dtiprocess = tools.DTIProcess(self.software_info['dtiprocess']['path'])
            # Determine which scalars need to be generated
            scalars_to_generate = []
            for scalar in ['FA', 'MD', 'AD', 'RD']:
                scalar_col_header = parameter_to_col_map[scalar]
                scalars_to_generate.append(scalar)
                df[scalar_col_header] = ''  # initialize the column as a string

            for index, row in df.iterrows():
                subject_id = str(row[parameter_to_col_map['Case ID']])
                path_to_original_dti_image = row[parameter_to_col_map['Original DTI Image']]
                scalar_img_folder_path = Path(output_base_dir).joinpath("scalar_images").joinpath(subject_id)
                output_stem = scalar_img_folder_path.joinpath(Path(path_to_original_dti_image).stem).__str__()
                # check if scalar_img_folder_path already exists
                if scalar_img_folder_path.exists() and not recompute_scalars:
                    logger(f"Skipping recomputation of scalars {', '. join(scalars_to_generate)} for subject " + subject_id)
                else:
                    scalar_img_folder_path.mkdir(parents=True, exist_ok=True)
                    options = ['--correction', 'none', '--saveScalarsAsFloat']
                    # run dtiprocess to generate scalar images
                    dtiprocess.measure_scalar_list(path_to_original_dti_image, output_stem, scalars_to_generate, options)

                # update the dataframe with the paths to the scalar images
                for scalar in scalars_to_generate:
                    scalar_col = parameter_to_col_map[scalar]
                    scalar_img_path_str = output_stem.__str__() + '_' + scalar + '.nrrd'
                    df.at[index, scalar_col] = scalar_img_path_str

        # write the modified dataframe to the output directory
        df.to_csv(Path(output_base_dir).joinpath(Path(path_to_csv).stem.__str__() + '_with_scalars.csv'), index=False)

        parameterized_fibers_path = Path(output_base_dir).joinpath('parameterized_fibers')
        parameterized_fibers_path.mkdir(parents=True, exist_ok=True)
        # iterate over the rows of the CSV
        for prop in properties_to_profile:
            logger(f"Extracting property {prop} from column header '{parameter_to_col_map[prop]}'")
            prop_output_path: Path = Path(output_base_dir).joinpath(prop)
            for tract in tracts:
                # create the directory for the output for the scalar property
                tract_name_stem: str = Path(tract).stem
                tract_output_path: Path = prop_output_path.joinpath(tract_name_stem)
                tract_output_path.mkdir(parents=True, exist_ok=True)
                logger(f"Extracting profile for tract {tract}")
                tract_absolute_filename = Path(atlas_path).joinpath(
                    tract)  # concatenate the atlas path with the tract name
                # Create dataframe to track statistics for this tract
                tract_stat_df: pd.DataFrame = None
                for row_index, row in df.iterrows():
                    subject_id = row.iloc[0]
                    # Find path to scalar image in the dataframe
                    scalar_img_path = row[parameter_to_col_map[prop]]
                    fiberprocess_output_path: str = tract_output_path.joinpath(
                        f'{subject_id}_' + tract.replace('_extracted_done', f'_{prop}_profile')).__str__()
                    fiberpostprocess_output_path: str = fiberprocess_output_path.__str__().replace('.vtk',
                                                                                                   '_processed.vtk')
                    dtitractstat_output_path: str = fiberpostprocess_output_path.replace('.vtk', '.fvp')
                    scalar_name = prop
                    if Path(fiberprocess_output_path).exists() and not recompute_scalars:
                        logger(f"Skipping fiberprocess of scalar {prop} for subject {subject_id}")
                    else:
                        # run fiberprocess
                        options = []
                        options += ['--scalarName', scalar_name]
                        options += ['--ScalarImage', scalar_img_path]
                        options += ['--no_warp']
                        if use_displacement_field:
                            options += ['--displacement_field', row[parameter_to_col_map['Deformation Field']]]
                        fiberprocess = tools.FiberProcess(self.software_info['fiberprocess']['path'])
                        fiberprocess.run(tract_absolute_filename.__str__(), fiberprocess_output_path,
                                         options=options)


                    if Path(fiberpostprocess_output_path).exists() and not recompute_scalars:
                        logger(f"Skipping fiberpostprocess of scalar {prop} for subject {subject_id}")
                    else:
                        # run fiberpostprocess
                        options = []
                        if mask is not None:
                            options += ['--mask', mask]
                        if noNaN:
                            options += ['--noNan']
                        fiberpostprocess = tools.FiberPostProcess(self.software_info['fiberpostprocess']['path'])
                        fiberpostprocess.run(fiberprocess_output_path.__str__(), fiberpostprocess_output_path, options=options)
                    if Path(dtitractstat_output_path).exists() and not recompute_scalars:
                        logger(f"Skipping dtitractstat of scalar {prop} for subject {subject_id}")
                    else:
                        # run dtitractstat
                        options = ['--parameter_list', prop, '--scalarName', prop]
                        if row_index == 0:
                            logger(f"Generating parameterized fiber profile for tract {tract}")
                            tract_name_stem: str = Path(tract).stem
                            parameterized_fiber_output_path: Path = Path(parameterized_fibers_path).joinpath(
                                tract_name_stem + "_parameterized.vtk")
                            if parameterized_fiber_output_path.exists() and not recompute_scalars:
                                logger(f"Skipping parameterized fiber generation of tract {tract}")
                            else:
                                logger(f"Generating parameterized fiber profile for tract {tract}")
                                tract_absolute_filename = Path(atlas_path).joinpath(
                                    tract)
                                options += ['-f', parameterized_fiber_output_path.__str__()]
                                options += ['--step_size', step_size]
                                options += ['--bandwidth', support_bandwidth]
                                options += ['--auto_plane_origin', plane_of_origin.lower()]
                                if noNaN:
                                    options += ['--remove_nan_fibers']
                        dtitractstat = tools.DTITractStat(self.software_info['dtitractstat']['path'])
                        dtitractstat.run(fiberpostprocess_output_path, dtitractstat_output_path, options=options)

                    # extract fvp data
                    fvp_data = pd.read_csv(dtitractstat_output_path, skiprows=[0, 1, 2, 3])

                    # write fvp data to csv
                    if tract_stat_df is None:
                        if result_case_columnwise:
                            tract_stat_df = pd.DataFrame(columns=["Arc Length"])
                            tract_stat_df["Arc Length"] = fvp_data["Arc_Length"].tolist()
                        else:
                            col_list = ['case_id'] + fvp_data["Arc_Length"].tolist()
                            tract_stat_df = pd.DataFrame(columns=col_list)

                    if result_case_columnwise:
                        new_col = fvp_data["Parameter_Value"].tolist()
                        tract_stat_df[subject_id] = new_col
                    else:
                        new_row_list = [subject_id] + fvp_data["Parameter_Value"].tolist()
                        tract_stat_df.loc[len(tract_stat_df)] = dict(zip(tract_stat_df.columns, new_row_list))

                tract_stat_df.to_csv(prop_output_path.joinpath(f'{tract_name_stem}_{prop}.csv'), index=False)

        self.result['output']['success'] = True
        return self.result








INFO = {
  "dtiplayground": { 
    "version" : "0.5.8b2"
  },
  "dmriplayground": {
    "version" : "0.5.8b2"
  },
  "dmriprep": {
    "version" : "0.5.8b2"
  },
  "dmriatlas": {
    "version" : "0.2.0b1"
  },
  "dmriautotract" : {
    "version" : "0.0.2b1"
  },
  "dmrifiberprofile" : {
    "version" : "0.5.8b2"
  }
}



import dtiplayground.dmri.preprocessing as prep
import dtiplayground.dmri.common as common
import yaml,os
from pathlib import Path
import dtiplayground.dmri.common.dwi as dwi 
import dtiplayground.dmri.common.tools as tools 

import sys

### Functions of Filter ###


### Function with two parameters : exclude gradients which are not in the range ###

def filter_bmin_bmax (b_min, b_max, image, inside=True):

    ## new code

    exclude_indexes = []
    grads = image.getGradients()
    for idx, g in enumerate(grads):
        if inside:
            if float(g['b_value']) > float(b_max) or float(g['b_value']) < float(b_min):
                exclude_indexes.append(idx)
        else:
            if float(g['b_value']) <= float(b_max) and float(g['b_value']) >= float(b_min):
                exclude_indexes.append(idx)

    return exclude_indexes


    # ## old code
    # liste=[]
    # for i in range (len(image.gradients)):
    #     liste.append(image.gradients[i].get('b_value'))
    #     #print("index : ",image.gradients[i].get('index'))
    #     #print("b_value : ",image.gradients[i].get('b_value'))
    # #image.gradients

    # new_list=[]
    # list2=liste
    # list_to_exclude=[]

    # logger("Filtering",prep.Color.INFO)
    # for j in range (len(image.gradients)):
    #     if image.gradients[j].get('b_value')<=b_min or image.gradients[j].get('b_value')>=b_max:
    #         #image.gradients.pop(j)
    #         #print("to exlcude : ", j)
    #         list2[j]=0
    #         new_list.append(j)
    #     #else :
        
    #         #list_to_exclude.append(j)
    #         #print("inside : ", j)
    # logger("Creating new gradient table",prep.Color.INFO)
    # for i in range(0,len(list2)):
    #     if list2[i]==0:
    #         #print("i : ",i)
    #         #print(image.gradients[i])
    #         image.gradients[i]={}
    #         #print(len(image.gradients))
    #         #image.gradients
    # logger("Image with new gradient table",prep.Color.INFO)
    # image.gradients= [i for i in image.gradients if i!={}]
    # for i in range(0, len(image.gradients)):
    #     #print(i)
    #     image.gradients[i][('index')]=i
    # #image.gradients

    # image.information['sizes'][3]=len(image.gradients)
    # logger("New len of gradient table : {}".format(len(image.gradients)),prep.Color.INFO)
    # #return image

### Function with two parameters : exclude gradients which are between bmin and bmax ###

# def exclude_between_bmin_bmax (b_min, b_max, image):
#     liste=[]
#     for i in range (len(image.gradients)):
#         liste.append(image.gradients[i].get('b_value'))
#         #print("index : ",image.gradients[i].get('index'))
#         #print("b_value : ",image.gradients[i].get('b_value'))
    
#     #image.gradients
    
#     new_list=[]
#     list2=liste
#     list_to_exclude=[]
    

#     logger("Filtering",prep.Color.INFO)
#     for j in range (len(image.gradients)):
#         if image.gradients[j].get('b_value')>=b_min and image.gradients[j].get('b_value')<=b_max:
#             #print("to exlcude : ", j)
#             list2[j]=0
#             new_list.append(j)
#         #else :
        
#             #list_to_exclude.append(j)
#             #print("inside : ", j)

#     logger("Creating new gradient table",prep.Color.INFO)
#     for i in range(0,len(list2)):
#         if list2[i]==0:
#             #print("i : ",i)
#             #print(image.gradients[i])
#             image.gradients[i]={}
#             #print(len(image.gradients))
#             #image.gradients
#     logger("Image with new gradient table",prep.Color.INFO)
#     image.gradients= [i for i in image.gradients if i!={}]
#     for i in range(0, len(image.gradients)):
#         #print(i)
#         image.gradients[i][('index')]=i
#     #image.gradients

#     image.information['sizes'][3]=len(image.gradients)
#     logger("New len of gradient table : {}".format(len(image.gradients)),prep.Color.INFO)
#     #return image

# ### Function with one parameter : exclude gradients which have b_values below this parameter b_min ###

# def filter_bmin(b_min, image):
#     liste=[]
#     for i in range (len(image.gradients)):
#         liste.append(image.gradients[i].get('b_value'))
#     #image.gradients
    
#     new_list=[]
#     list2=liste
#     list_to_exclude=[]
#     logger("Filtering",prep.Color.INFO)
#     for j in range (len(image.gradients)):
#         if image.gradients[j].get('b_value')<=b_min:
#             #print("to exlcude : ", j)
#             list2[j]=0
#             new_list.append(j)
#         #else :
#         #    print("inside : ", j)
#     logger("Creating new gradient table",prep.Color.INFO)      
#     for i in range(0,len(list2)):
#         if list2[i]==0:
#             image.gradients[i]={}
#     logger("Image with new gradient table",prep.Color.INFO)
#     image.gradients= [i for i in image.gradients if i!={}]
#     for i in range(0, len(image.gradients)):
#         image.gradients[i][('index')]=i

#     image.information['sizes'][3]=len(image.gradients)
#     logger("New len of gradient table : {}".format(len(image.gradients)),prep.Color.INFO)
#     #return image


### Function with one parameter : exclude gradients which have b_values above this parameter b_max ###

# def filter_bmax(b_max, image):
#     liste=[]
#     for i in range (len(image.gradients)):
#         liste.append(image.gradients[i].get('b_value'))
#     #image.gradients
    
#     new_list=[]
#     list2=liste
#     list_to_exclude=[]
#     logger("Filtering",prep.Color.INFO)
#     for j in range (len(image.gradients)):
#         if image.gradients[j].get('b_value')>=b_max:
#             #print("to exlcude : ", j)
#             list2[j]=0
#             new_list.append(j)
#         #else :
#         #    print("inside : ", j)
#     logger("Creating new gradient table",prep.Color.INFO)   
#     for i in range(0,len(list2)):
#         if list2[i]==0:
#             image.gradients[i]={}
#     logger("Image with new gradient table",prep.Color.INFO)
#     image.gradients= [i for i in image.gradients if i!={}]
#     for i in range(0, len(image.gradients)):
#         image.gradients[i][('index')]=i

#     image.information['sizes'][3]=len(image.gradients)
#     logger("New len of gradient table : {}".format(len(image.gradients)),prep.Color.INFO)
    #return image


class IMAGE_Filter(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        global logger
        logger = self.logger.write
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        
        # << TODOS>>
        gradient_indexes_to_remove=[]

        logger("Choosing the b-value thresholding mode : ", prep.Color.INFO)
        if self.protocol['b_lower']<self.protocol['b_upper']:
            if self.protocol['tresholding_mode']=='one_treshold_below':
                gradient_indexes_to_remove= filter_bmin_bmax(b_min=0, b_max=self.protocol['b_tresh'], image=self.image, inside=True)
            elif self.protocol['tresholding_mode']=='one_treshold_above':
                gradient_indexes_to_remove= filter_bmin_bmax(b_min=self.protocol['b_tresh'],b_max=10e9, image=self.image, inside=True)
            elif self.protocol['tresholding_mode']=='two_tresholds_within':
                gradient_indexes_to_remove= filter_bmin_bmax(b_min=self.protocol['b_lower'], b_max=self.protocol['b_upper'], image=self.image, inside=True)
            elif self.protocol['tresholding_mode']=='two_tresholds_outside':
                gradient_indexes_to_remove= filter_bmin_bmax(b_min=self.protocol['b_lower'], b_max=self.protocol['b_upper'], image=self.image, inside=False)
        elif self.protocol['b_lower']>self.protocol['b_upper']:
            logger("Error : b_lower must be lower than b_upper or b_upper must be higher than b_lower", prep.Color.ERROR)
            sys.exit()
        elif self.protocol['b_lower']<0:
            logger("Error : b_lower must be positive", prep.Color.ERROR)
            sys.exit()
        elif self.protocol['b_upper']<0:
            logger("Error : b_upper must be positive", prep.Color.ERROR)
            sys.exit()
        elif self.protocol['b_tresh']<0:
            logger("Error : b_tresh must be positive", prep.Color.ERROR)
            sys.exit()
        logger("Treatment done", prep.Color.INFO)

        self.result['output']['excluded_gradients_original_indexes']=self.image.convertToOriginalGradientIndex(gradient_indexes_to_remove)
        self.result['output']['success']=True
        return self.result




##################################################################################
### Module : filebrowser.py
### Description : Remote(Local) File browser 
###
###
###
### Written by : scalphunter@gmail.com ,  2022/10/07
### Copyrights reserved by NIRAL
##################################################################################

import sys, traceback

### API
from flask import request, Response , send_from_directory, jsonify
import json
from . import utils
from pathlib import Path
import multiprocessing as mp
mp.set_start_method("fork")
from dtiplayground.config import INFO 
import dtiplayground.dmri.common.dwi as dwi
import cv2
import io
import numpy as np 

class ApplicationAPI:
    def __init__(self,server,**kwargs):
        self.server = server
        self.app=self.server.app
        self.filecache={}
        self.initEndpoints()

##### Endpoints

    def initEndpoints(self):
        @self.app.route('/api/v1/version',methods=['GET'])
        def _checkAppVersion():
            sc=200
            res=None
            request_id=utils.get_request_id()
            try:
                res= self.checkAppVersion()
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

        @self.app.route('/api/v1/app',methods=['GET'])
        def _getUserInfo():
            sc=200
            res=None
            request_id=utils.get_request_id()
            try:
                res= self.getAppInfo()
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        @self.app.route('/api/v1/files',methods=['GET'])
        def _getFileList():
            sc=200
            res=None
            param_rootdir=request.args.get('root_dir',default='/',type=str)
            request_id=utils.get_request_id()
            try:
                res= self.browseDirectory(param_rootdir)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        @self.app.route('/api/v1/file-url',methods=['GET'])
        def _getFileUrl():
            sc=200
            res=None
            param_path=request.args.get('path',type=str)
            request_id=utils.get_request_id()
            try:
                url = request.url
                url = url.replace('/api/v1/file-url','/api/v1/download')
                res= { 'download_url' : url }
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        @self.app.route('/api/v1/download',methods=['GET'])
        def _getDownload():
            print(request.url)
            sc=200
            res={}
            param_path=request.args.get('path',type=str)
            request_id=utils.get_request_id()
            try:
                path=Path(param_path)
                filename = path.name
                destDir = path.parent
                if path.exists():
                    return send_from_directory(str(destDir), filename, as_attachment=True)
                else:
                    raise Exception("There is no such file")
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)

        @self.app.route('/api/v1/files/get-text',methods=['GET'])
        def _getFileTextContent():
            sc=200
            res=None
            param_path=request.args.get('path',default='/',type=str)
            param_last_line=request.args.get('last_line',default=0,type=int)
            request_id=utils.get_request_id()
            try:
                res= self.getTextFileContentAsArray(param_path,param_last_line)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        @self.app.route('/api/v1/files/get-text-whole',methods=['GET'])
        def _getFileWholeContent():
            sc=200
            res=None
            param_path=request.args.get('path',default='/',type=str)
            request_id=utils.get_request_id()
            try:
                res= self.getTextFileContentAsWhole(param_path)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        @self.app.route('/api/v1/files/get-readme',methods=['GET'])
        def _getReadMe():
            sc=200
            res=None
            request_id=utils.get_request_id()
            try:
                import dtiplayground
                path = Path(dtiplayground.__file__).resolve().parent.parent.joinpath('README.md');
                res= self.getTextFileContentAsWhole(path)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        ###### DWI image
        @self.app.route('/api/v1/files/load',methods=['GET'])
        def _loadFileAsCache():
            sc=200
            res=None
            request_id=utils.get_request_id()
            param_path=request.args.get('filename',default=None,type=str)
            param_key=request.args.get('key', default='image', type=str)
            if param_path is None: raise Exception('No file selected')
            try:
                fn = Path(param_path).resolve()
                res= self.loadFileAsCache(fn, param_key)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp        

        ####### Image browsing (DWI)
        @self.app.route('/api/v1/dwi/<img_id>/<grad_idx>/<axis_idx>/<slice_idx>',methods=['GET'])
        def _getFrameFromDWI(img_id,grad_idx,axis_idx,slice_idx): ## img_id is not used, just to prevent browser cacheing
            res=None
            sc=500 
            data=None
            grad_idx=int(grad_idx)
            axis_idx=int(axis_idx)
            slice_idx=int(slice_idx)
            param_min = request.args.get('min',default=0,type=int)
            param_max = request.args.get('max',default=10e6,type=int)
            try:
                # if axis_idx == 0:
                #     res = self.filecache['dwi'].images[int(slice_idx),:,:,int(grad_idx)]
                # elif axis_idx == 1:
                #     res = self.filecache['dwi'].images[:,int(slice_idx),:,int(grad_idx)]
                # elif axis_idx == 2:
                #     res = self.filecache['dwi'].images[:,:,int(slice_idx),int(grad_idx)]
                # else: raise Exception('No such axis')

                # out = (res >= param_min) * res
                # out[out >= param_max] = param_max
                out = self.filecache['dwi'].getImageSlice4D(axis_idx,slice_idx,grad_idx,normalized=True, display_range=[param_min, param_max])
                ok,res=cv2.imencode('.jpeg',out, [int(cv2.IMWRITE_JPEG_QUALITY), 50]) ### compress 
                if ok:                       
                    data=io.BytesIO(res).read() 
                else:
                    raise Exception("Failed to encode")
                sc=200
            except Exception as e:
                msg=traceback.format_exc()
                err_msg="{}:{}".format(str(e),msg)
                print(str(e))
                print(msg)
            finally:
                resp=Response(data,status=sc)
                resp.headers['Content-Type']='application/octet-stream'
                resp.headers['Image-Format']='jpeg'
                return resp

        ####### Multiprocessing

        @self.app.route('/api/v1/process', methods=['GET'])
        def _get_processes():
            sc=200
            res=None
            request_id=utils.get_request_id()
            try:
                res= self.getProcesses()
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        @self.app.route('/api/v1/process/<pid>', methods=['DELETE'])
        def _kill_processes(pid):
            sc=200
            res=None
            request_id=utils.get_request_id()
            try:
                res= self.killProcess(pid)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

    def loadFileAsCache(self, filename, filekey):
        meta={}
        if filekey.lower() == 'dwi':
            #load DWI and put it in cache 
            self.filecache[filekey] = dwi.DWI(str(filename))
            meta = {
                'info': self.filecache[filekey].information,
                'gradients': self.filecache[filekey].getGradients()
            }
            del meta['info']['thicknesses']

        out = {
            'filename': str(filename),
            'type': filekey,
            'meta': meta,
        }
        return out

    #### image browsing

    #### Process

    def getProcesses(self):
        res = mp.active_children()
        res = list(map(lambda x: {'pid': x.pid, 'name' : x.name }, res))

        return res

    def killProcess(self, _id): # id can bd pid or execution id
        allprocs = mp.active_children()
        tokill = filter(lambda x: str(x.pid) == _id or x.name == _id,allprocs)
        for p in tokill:
            p.kill()

        res = list(map(lambda x: {'pid': x.pid, 'name' : x.name }, tokill))
        return res

    #### Filesystem DMRIPlayground
    def checkAppVersion(self):
        return { 'version' : "0.0.1" }

    def browseDirectory(self, rootdir):
        p = Path(rootdir).glob("*")
        parent = Path(rootdir).parent.__str__()
        fixed_paths = [
            {
                'path': parent,
                'is_dir': True,
                'name' : '..',
                'is_real': False
            }
        ]
        paths = ({'path': str(x), 'is_dir': x.is_dir(), "name": x.name, 'is_real': True } for x in p)
        paths = list(paths)
        paths.sort(key=lambda x: (not x['is_dir'], x['name']))
        return fixed_paths + list(paths) 

    def getTextFileContentAsArray(self, path, lastline):
        p = Path(path)
        with open(p,'r') as f:
            content=f.readlines()
        return content[lastline:-1]

    def getTextFileContentAsWhole(self, path):
        p = Path(path)
        with open(p,'r') as f:
            content=f.read()
        return content

    def getAppInfo(self):
        import socket
        res = {
            'version' : INFO['dtiplayground']['version'],
            'home_dir': Path.home().__str__(),
            'config_dir' : str(self.server.config_dir),
            'hostname': socket.gethostname()
        }
        return res



#!python

import shutil
import os 
import importlib
from pathlib import Path
import traceback
import time
import yaml
import uuid
import sys
import re
from functools import wraps
import subprocess

from dtiplayground.config import INFO as info
import dtiplayground
import dtiplayground.dmri.common as common
import dtiplayground.dmri.common.module as module

logger=common.logger.write 
color= common.Color

class AppBase:
    def __init__(self,config_root, app_name,*args,**kwargs):
        self.config_dir = Path(config_root)
        self.app = {}
        self.app['application'] = app_name
        self.getAppInfo()
        if 'application_dir' in self.app:
            self.log_file = str(Path(self.app['application_dir']).joinpath('log.txt'))
            self.app.setdefault('no_log_timestamp',False)
            self.app.setdefault('log', self.log_file)
            self.app.setdefault('no_verbosity', False)
            self.app.setdefault('execution_id',  common.get_uuid())
            init_fn = Path(self.app['application_dir']).joinpath('init.yml')
            globalvar_fn =  Path(self.config_dir).joinpath('global_variables.yml')
            if globalvar_fn.exists():
                self.app['parameters'] = yaml.safe_load(open(globalvar_fn,'r'))
                pass
            self.kwargs=kwargs
        else:
            self.kwargs=kwargs
        self.globalvars = {}
        logger(yaml.dump(self.app))

    def getAppInfo(self):
        version=info[self.app['application']]['version']
        application_dir=Path(self.config_dir).joinpath('{}-{}'.format(self.app['application'],str(version))).resolve()
        res = {
            'version' : version,
            'config_dir' : str(self.config_dir),
            'application_dir' : str(application_dir)
        }
        self.app.update(res)
        self.getAppInfoImpl()

        return self.app 

    def getAppInfoImpl(self):
        return self.app

    def initialize(self, options):
        ## reparametrization
        _args = {
            'log' : self.app['log'],
            'execution_id' : self.app['execution_id'],
            'no_verbosity' : self.app['no_verbosity'],
            'no_log_timestamp' : self.app['no_log_timestamp']
        }
        _args.setdefault('tools_dir',None)
        _args.setdefault('install_tools',False)

        home_dir = Path(self.app['application_dir'])
        version = self.app['version']

        ## Function begins
        home_dir.mkdir(parents=True,exist_ok=True)
        tools_dir = Path(self.app['config_dir'])
        if _args['tools_dir'] is not None:
            tools_dir = Path(_args['tools_dir'])
        else:
            tools_dir = Path(self.app['config_dir'])

        globalvars_fn=Path(self.app['config_dir']).joinpath('global_variables.yml')

        self.globalvars = {}
        self.globalvars.setdefault('dtiplayground-tools',{'info':{},'path':None})
        self.globalvars.setdefault('fsl',{'info':{},'path':None})

        if globalvars_fn.exists():
            self.globalvars = yaml.safe_load(open(globalvars_fn,'r'))

        if tools_dir is not None:
            if tools_dir.exists():
                dptdir = tools_dir.joinpath('dtiplayground-tools')
                fsldir = tools_dir.joinpath('FSL')

                tmpdir=self._get_dpg_dir()
                if  tmpdir is not None:
                    dptdir = Path(tmpdir)
                    if tmpdir != self.globalvars['dtiplayground-tools']['path']:
                        logger("Use the installed DTIPlaygroundTools? [Y/n]")
                        yn=input()
                        if yn.lower() == 'n':
                            pass
                        else:
                            self.globalvars['dtiplayground-tools']['path'] = dptdir.__str__()

                tmpdir=self._get_fsl_dir()
                if  tmpdir is not None:
                    fsldir = Path(tmpdir)
                    if tmpdir != self.globalvars['fsl']['path']:
                        logger("Use the installed FSL? [Y/n]")
                        yn=input()
                        if yn.lower() == 'n':
                            pass
                        else:
                            self.globalvars['fsl']['path'] = fsldir.__str__()

                if self.globalvars['dtiplayground-tools']['path'] is not None:
                    dptdir = Path(self.globalvars['dtiplayground-tools']['path'])
                if self.globalvars['fsl']['path'] is not None:
                    fsldir = Path(self.globalvars['fsl']['path'])

                ## DTIPlayground Tools


                if not dptdir.joinpath('info.yml').exists():
                    if not _args['install_tools']:
                        logger('Couldn\'t find the DTIPlayground tools at: {}'.format(dptdir.__str__()),color.ERROR)
                        logger('Do you want to install the DTIPlaygroundTools now? [Y/n]')
                        yn = input()
                        if yn=='n':
                            logger('Please input the directory where DTIPlaygroundTools is installed:')
                            dptdir = input()
                            if len(dptdir) < 1:
                                logger('Initialization aborted',color.ERROR)
                                exit(1)
                            else:
                                dptdir = Path(dptdir).resolve()
                                if Path(dptdir).joinpath('info.yml').exists():
                                    pass
                                else:
                                    logger('Couldn\' find the DTIPlaygroundTools. Initialization aborted',color.ERROR)
                                    exit(1)
                        else:
                            logger('Installation Directory [~/.niral-dti/]:')
                            dptdir = input()
                            if len(dptdir) < 1:
                                dptdir = Path(tools_dir)
                            else:
                                dptdir=Path(dptdir)
                            dptdir=dptdir.expanduser()
                            dptdir=Path(os.path.expandvars(dptdir))
                            logger('Installing to : {}'.format(dptdir.__str__()),color.PROCESS)
                            self.install_tools({'output_dir': Path(dptdir).resolve(),'no_fsl':True})
                            dptdir = dptdir.joinpath('dtiplayground-tools')
                    else:
                        logger('Installation Directory [~/.niral-dti/]:')
                        dptdir = input()
                        if len(dptdir) < 1:
                            dptdir = Path(tools_dir)
                        else:
                            dptdir=Path(dptdir)
                        dptdir=dptdir.expanduser()
                        dptdir=Path(os.path.expandvars(dptdir))
                        logger('Installing to : {}'.format(dptdir.__str__()),color.PROCESS)
                        self.install_tools({'output_dir': Path(dptdir).resolve(), 'no_fsl':True})
                        dptdir = dptdir.joinpath('dtiplayground-tools')

                ## FSL
                os.chdir(Path.home())
                if not fsldir.joinpath('etc/fslversion').exists():
                    if not _args['install_tools']:
                        logger('Couldn\'t find the FSL at: {}'.format(fsldir.__str__()),color.ERROR)
                        logger('Do you want to install the FSL now? [Y/n]')
                        yn = input()
                        if yn=='n':
                            logger('Please input the directory where FSL is installed:')
                            fsldir = input()
                            if fsldir is None:
                                logger('Initialization aborted',color.ERROR)
                                exit(1)
                            else:
                                fsldir = Path(fsldir).resolve()
                                if Path(fsldir).joinpath('etc/fslversion').exists():
                                    pass
                                else:
                                    logger('Couldn\' find the FSL. Initialization aborted',color.ERROR)
                                    exit(1)
                        else:
                            logger('Installation Directory [~/.niral-dti/]:')
                            fsldir = input()
                            if len(fsldir) < 1:
                                fsldir = tools_dir
                            else:
                                fsldir=Path(fsldir)
                            fsldir=fsldir.expanduser()
                            fsldir=Path(os.path.expandvars(fsldir))
                            logger('Installing to : {}'.format(fsldir.__str__()),color.PROCESS)
                            self.install_tools({'output_dir': Path(fsldir).resolve(),'no_dtiplayground_tools':True})
                            fsldir=fsldir.joinpath('FSL')
                    else:
                        logger('Installation Directory [~/.niral-dti/]:')
                        fsldir = input()
                        if len(fsldir) < 1:
                            fsldir = tools_dir
                        else:
                            fsldir=Path(fsldir)
                        fsldir=fsldir.expanduser()
                        fsldir=Path(os.path.expandvars(fsldir))
                        logger('Installing to : {}'.format(fsldir.__str__()),color.PROCESS)
                        self.install_tools({'output_dir': Path(fsldir).resolve(), 'no_dtiplayground_tools':True})
                        fsldir=fsldir.joinpath('FSL')

                if self.globalvars is None:
                    self.globalvars={}
                self.globalvars.update({
                    "dtiplayground-tools" :{
                        "path" : dptdir.resolve().__str__(),
                        "info" : yaml.safe_load(open(dptdir.joinpath('info.yml'),'r'))
                    },
                    "fsl" : {
                        "path" : fsldir.resolve().__str__(),
                        "info" : {
                            'name' : 'fsl',
                            "verson" : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
                        }
                    }
                })
                yaml.dump(self.globalvars, open(globalvars_fn,'w'))
            else:
                logger("Couldn't find tool directory {}".format(str(tools_dir)),color.ERROR)
                exit(1)

        yaml.dump(self.globalvars, open(globalvars_fn,'w'))

        self.app['global_variables'] = self.globalvars
        self.initialize_logger(_args)
        self.initializeImpl(options)
        initialization_fn = Path(self.app['application_dir']).joinpath("init.yml")
        yaml.dump(self.app, open(initialization_fn,'w'))
        logger("Initialized. Local configuration will be stored in {}".format(str(home_dir)),color.OK)



    def install_tools(self,_options={}):
        # @self.after_initialized
        def _install_tools(options):
            ## reparametrization
            options.setdefault('no_fsl',False)
            options.setdefault('no_dtiplayground_tools',False)
            options.setdefault('install_only', True)
            options.setdefault('build', False)
            options.setdefault('no_remove', False)
            options.setdefault('clean_install', False)
            options.setdefault
            args = {
                'output_dir' : options['output_dir'],
                'clean_install': options['clean_install'],
                'no_fsl' : options['no_fsl'],
                'no_dtiplayground_tools': options['no_dtiplayground_tools'],
                'install_only' : options['install_only'],
                'build' : options['build'],
                'no_remove' : options['no_remove'],
            }
            args['output_dir']=os.path.expandvars(args['output_dir'])
            logger('Installing to : {} [Y/n]'.format(Path(args['output_dir'])))
            yn = input()
            if yn.lower() == 'n':
                logger('Specify the installation directory:')
                inst_dir = input()
                if len(inst_dir) < 1:
                    logger('Installation aborted',color.ERROR)
                    exit(1)
                else:
                    inst_dir = Path(inst_dir)
                    inst_dir = inst_dir.expanduser()
                    inst_dir = Path(os.path.expandvars(inst_dir))
                    args['output_dir'] = inst_dir
                    logger('Installing tools to {}...'.format(args['output_dir']),color.PROCESS)
            else:
                pass

            home_dir=Path(self.app['application_dir'])
            home_dir.parent.mkdir(parents=True,exist_ok=True)
            # globalvars_fn = Path(os.path.expandvars("$HOME/.niral-dti/global_variables.yml"))
            globalvars_fn = home_dir.parent.joinpath('global_variables.yml')
            Path(globalvars_fn).parent.mkdir(parents=True,exist_ok=True)
            globalvars = {}
            if globalvars_fn.exists():
                globalvars = yaml.safe_load(open(globalvars_fn,'r'))

            rootdir = Path(os.path.expandvars(args['output_dir'])).resolve()
            rootdir.mkdir(parents=True,exist_ok=True)
            tempdir = rootdir.joinpath('temp-dtiplayground-tools')
            outputdir = rootdir.joinpath('dtiplayground-tools')
            fsldir = rootdir.joinpath('FSL')
            clean_install = args['clean_install']
            leave_files_after_installation = args['no_remove']
            no_fsl = args['no_fsl']
            no_dtiplayground_tools = args['no_dtiplayground_tools']
            install_only = args['install_only']
            build=args['build']
            ### Clean install preparations
            if clean_install:
                logger("Removing temporary, existing packages (dtiplayground-tools only) directories ...")
                if tempdir.exists():
                    shutil.rmtree(tempdir)
                if outputdir.exists():
                    shutil.rmtree(outputdir)
                logger("Removing files finished")

            ### FSL


            if not no_fsl and not fsldir.joinpath('etc/fslversion').exists():
                import dtiplayground.dmri.common.data

                fslinstaller_fn=Path(dtiplayground.dmri.common.data.__file__).parent.joinpath('fslinstaller.py')
                command=['python', fslinstaller_fn ,'-d',fsldir.resolve().__str__()] 
                subprocess.run(command)
                info = {
                    'name' : 'fsl',
                    'version' : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
                }
                globalvars['fsl']={
                    'path' : fsldir.resolve().__str__(),
                    'info' : info
                }
                yaml.dump(info, open(fsldir.joinpath('info.yml'),'w'))
            else:
                if fsldir.exists():
                    info = {
                        'name': 'fsl',
                        'version' : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
                    }
                    yaml.dump(info, open(fsldir.joinpath('info.yml'),'w'))
                    globalvars['fsl']={
                        'path' : fsldir.resolve().__str__(),
                        'info' : info
                    }            

            ### DTI playground tools (build with docker, centos7 for now)
            tempdir.mkdir(parents=True,exist_ok=True)
            os.chdir(tempdir)
            srcdir = tempdir.joinpath('dtiplaygroundtools')
            info={}


            if not outputdir.joinpath('info.yml').exists() and not no_dtiplayground_tools:

                if build:
                    if not srcdir.exists():
                        logger("Fetching source code ...")
                        fetch_dtiplaygroundtools = ["git","clone","https://github.com/niraluser/dtiplaygroundtools.git"]
                        subprocess.run(fetch_dtiplaygroundtools)
                        logger("Source codes downloaded")
                    
                    os.chdir(srcdir.joinpath('dockerfiles'))
                    build_command = ['./build.sh']
                    logger("Building software packages ... ")
                    subprocess.run(build_command)
                    tar_filename = srcdir.joinpath('dist/dtiplayground-tools.tar.gz')
                    untar_command = ['tar','xvfz',tar_filename.resolve().__str__(), '-C', rootdir.resolve().__str__()]
                    logger("Installing softwares to the output directory {}".format(outputdir.resolve().__str__()))
                    subprocess.run(untar_command)
                    ### read version info
                    info_fn = outputdir.joinpath('info.yml')
                    if info_fn.exists():
                        info = yaml.safe_load(open(info_fn,'r'))
                else: # prebuilt package (default)
                    tar_filename = outputdir.joinpath('dtiplayground-tools.tar.gz')
                    if not tar_filename.exists():
                        remote_package = "https://github.com/NIRALUser/DTIPlaygroundTools/releases/download/v0.0.1/dtiplayground-tools.tar.gz"
                        fetch_package = ['wget',remote_package,'-P',outputdir.resolve().__str__()]
                        subprocess.run(fetch_package)
                    untar_command = ['tar','xvfz',tar_filename.resolve().__str__(), '-C', rootdir.resolve().__str__()]
                    logger("Installing softwares to the output directory {}".format(outputdir.resolve().__str__()))
                    subprocess.run(untar_command)
                    info_fn = outputdir.joinpath('info.yml')
                    if info_fn.exists():
                        info = yaml.safe_load(open(info_fn,'r'))
                    logger("Removing temporary files ...")
                    tar_filename.unlink()

            else:
                info = yaml.safe_load(open(outputdir.joinpath('info.yml'),'r'))

            globalvars['dtiplayground-tools']={
                'path' : outputdir.resolve().__str__(),
                'info' : info
            }

            if not leave_files_after_installation:
                logger("Removing temporary files ...")
                shutil.rmtree(tempdir)
                logger("Removing temporary files finished")

            logger("Saving tool information to {}".format(globalvars_fn.__str__()))
            yaml.dump(globalvars, open(globalvars_fn,'w'))
            if home_dir.joinpath('software_paths').exists() and not install_only:
                self._update_software_paths(globalvars)
            logger("Installation completed")
            return True
        return _install_tools(_options)

    def initializeImpl(self, options):
        pass

    def run(self,options): ### options must have output_dir 
        return self.runImpl(options)

    def runImpl(self, options):
        return True

    ################################################
    ## decorators (only used in the memeber function)
    #################################################

    def after_initialized(self,func): #decorator for other functions other than command_init
        def wrapper(args):
            home_dir=Path(self.app['application_dir'])
            if home_dir.exists() and home_dir.joinpath('init.yml').exists() and home_dir.parent.joinpath('global_variables.yml').exists():
                logger("Configuration directory is found : {}".format(self.app['application_dir']),color.OK)
                config_file=home_dir.joinpath('config.yml')
                environment_file=home_dir.joinpath('environment.yml')
                self.initialize_logger(self.app)
                return func(args)
            else: 
                # raise Exception("Not initialized, please run init command at the first use")
                logger("Configuration directory is not found or some files are missing, commencing initialization ...",color.WARNING)
                self.initialize(args)
                config_file=home_dir.joinpath('config.yml')
                environment_file=home_dir.joinpath('environment.yml')
                self.initialize_logger(self.app)
                return func(args)
        return wrapper


    #################################################
    ######## utilities
    #################################################

    def initialize_logger(self,_args):
        ## default log setting
        common.logger.setLogfile(_args['log'])
        common.logger.setTimestamp(not _args['no_log_timestamp'])
        common.logger.setVerbosity(not _args['no_verbosity'])


    def _update_software_paths(self, globalvars):
        home_dir=Path(self.app['application_dir'])
        software_info_path=Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/common/data/software_paths.yml")
        software_paths=yaml.safe_load(open(software_info_path,'r'))
        software_paths=self._resolve_softwarepaths(software_paths, globalvars)
        software_filename=home_dir.joinpath('software_paths.yml')
        yaml.dump(software_paths, open(software_filename,'w'))
    
    def _get_fsl_dir(self):
        logger('Looking up FSLDIR environment variable...', color.PROCESS)
        if 'FSLDIR' in os.environ:
            fsldir = os.environ['FSLDIR']
            logger('Found FSLDIR: {}'.format(fsldir),color.OK)
            if Path(fsldir).joinpath('etc/fslversion').exists():
                return fsldir
            else:
                logger('Error : {}/etc/fslversion doesn\'t exist, installation might be required'.format(fsldir),color.ERROR)
                return None

    def _get_dpg_dir(self):
        logger('Looking up DTIPLAYGROUNDTOOLS environment variable...', color.PROCESS)
        if 'DTIPLAYGROUNDTOOLS' in os.environ:
            dptdir = os.environ['DTIPLAYGROUNDTOOLS']
            logger('Found DTIPLAYGROUNDTOOLS: {}'.format(dptdir),color.OK)
            if Path(dptdir).joinpath('info.yml').exists():
                return dptdir
            else:
                logger('Error : {}/info.yml doesn\'t exist, installation might be required'.format(dptdir),color.ERROR)
                return None     

    def _resolve_softwarepaths(self,spathobj, globalvars):
        if 'dtiplayground-tools' in globalvars:
            p = Path(globalvars['dtiplayground-tools']['path'])
            if p.exists():
                os.environ['DTIPLAYGROUNDTOOLS'] = p.resolve().__str__()

        if 'fsl' in globalvars:
            p = Path(globalvars['fsl']['path'])
            if p.exists():
                os.environ['FSL'] = p.resolve().__str__()

        if  'DTIPLAYGROUNDTOOLS' not in os.environ:
            os.environ['DTIPLAYGROUNDTOOLS']=os.path.expandvars("$HOME/.niral/dtiplayground-tools")
            tooldir=os.environ['DTIPLAYGROUNDTOOLS']
            if Path(tooldir).exists():
                logger("DTI Playground tools directory found at {}".format(tooldir), color.OK)
            else:
                logger("DTI Playground tools directory is set to {}".format(tooldir), color.WARNING)
                logger("DTI Playground tools directory not found, please install toolkits or specify software paths manually", color.WARNING)
        else:
            tooldir=os.environ['DTIPLAYGROUNDTOOLS']
            if Path(os.environ['DTIPLAYGROUNDTOOLS']).exists():
                
                logger("DTI playground tools are found in {}".format(tooldir),color.OK)
            else:
                logger("DTI playground tools are not found in {}".format(tooldir),color.ERROR)
        if 'FSL' not in os.environ:
            logger("FSL directory is not set in environment variable FSL",color.WARNING)
            logger("Make sure FSL directory is set in the software paths",color.WARNING)
        else:
            if Path(os.environ['FSL']).exists():
                logger("FSL directory is found at {}".format(os.environ['FSL']),color.OK)
            else:
                logger("FSL directory is set in environment variable FSL, but doesn't exist",color.WARNING)
        
        sp = spathobj['softwares']
        for k,v in sp.items():
            resolved_path = os.path.expandvars(v['path'])
            sp[k]['path'] = resolved_path
        spathobj['softwares'] = sp
        return spathobj





from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class FiberProcess(ExternalToolWrapper):
    def __init__(self,binary_path= None, **kwargs):
        super().__init__(binary_path, **kwargs)
        self.binary_path=None
        if binary_path is not None:
            self.binary_path=binary_path
        elif 'softwares' in kwargs:
            self.binary_path=kwargs['softwares']['fiberprocess']['path']

    """
    FiberProcess Command Line Tool

    Usage:
      fiberprocess [options]

    Options:
      # I/O
      --fiber_file, -fiberFile           Input DTI fiber file. (Required)
      --fiber_output, -o                 Output fiber file. May be warped or updated with new data. (Required)
      --tensor_volume, -T                Input tensor volume. Interpolate tensor values from this field. (tensor or scalar required) 
      --scalarImage, -S                  Input scalar image file. Samples scalar values at fiber locations. (tensor or scalar required) 
      --scalarName                        Name for the pointData field in the output fiber file.
      --h_field, -H                      Input HField for warp and statistics lookup. (alternative to displacement field)
      --displacement_field, -D           Input Displacement Field for warp and statistics lookup. (usually needed to move images into atlas space)

      # Options
      --saveProperties, -p               Save tensor property as scalar data into the VTK (for VTK fiber files).
      --no_warp, -n                      Do not warp the geometry of tensors, only obtain new statistics.
      --fiber_radius, -R                 Set the radius of all fibers to this value.
      --index_space, -indexSpace         Use index-space for fiber output coordinates; otherwise, use world space.

      # Voxelize
      --voxelize, -V                     Voxelize fiber into a label map. Deformation is applied before voxelization.
      --voxelize_count_fibers           Count the number of fibers per voxel instead of just setting to 1.
      --voxel_label                       Label for voxelized fiber (default is 1).

      # Advanced Options
      --verbose, -v                      Produce verbose output.
      --noDataChange                     Do not change data.
    """
    """
    Calls the FiberProcess binary using the Python wrapper in the common tools. Uses arguments described above.
    """
    def run(self, fiber_file, output_path, options=[]):
        arguments=[]
        arguments+=['--fiber_file',fiber_file]
        arguments+=['--fiber_output', output_path]
        arguments+=options
        self.setArguments(arguments)
        return self.execute(arguments)




##################################################################################
### Module : dmriatlasbuilder.py
### Description : Atlasbuilder api
###
###
###
### Written by : scalphunter@gmail.com ,  2022/10/08
### Copyrights reserved by NIRAL
##################################################################################

import sys, traceback

### API
from flask import request, Response , send_from_directory, jsonify
import json
from . import utils
from pathlib import Path
import yaml
import multiprocessing
from multiprocessing import Process

class DMRIAtlasbuilderAPI:
    def __init__(self,server,**kwargs):
        self.server = server
        self.app=self.server.app
        self.initEndpoints()

##### Endpoints

    def initEndpoints(self):
        @self.app.route('/api/v1/dmriatlasbuilder/parameters',methods=['POST'])
        def _post_dmriab_params():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            try:
                req=request.get_json()
                res= self.generate_parameter_dir(req)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

        @self.app.route('/api/v1/dmriatlasbuilder',methods=['POST'])
        def _execute_dmriab_params():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            try:
                req=request.get_json()
                output_dir = req['output_dir']
                res= self.build_atlas(output_dir)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

        @self.app.route('/api/v1/dmriatlasbuilder',methods=['GET'])
        def _get_dmriab_template():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            try:
                res= self.getUITemplate()
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  


    def generate_parameter_dir(self,req):
        params = {
            'output_dir' : req['output_dir'],
            'hbuild' : req['hbuild'],
            'config' : req['config'],
            'greedy' : req['greedy']
        }
        output_dir=Path(params['output_dir'])
        output_dir.mkdir(exist_ok=True,parents=False)
        common_dir = output_dir.joinpath('common')
        common_dir.mkdir(exist_ok=True)

        config_path = common_dir.joinpath('config.yml')
        hbuild_path = common_dir.joinpath('h-build.yml')
        greedy_path = common_dir.joinpath('greedy.yml')

        config_json=output_dir.joinpath('common/config.json')
        hbuild_json=output_dir.joinpath('common/h-build.json')
        greedy_json=output_dir.joinpath('common/greedy.json')

        config = params['config']
        hbuild = params['hbuild']
        greedy = params['greedy']

        ## save config,hbuild
        yaml.safe_dump(config, open(config_path,'w'))
        yaml.safe_dump(hbuild, open(hbuild_path,'w'))
        yaml.safe_dump(greedy, open(greedy_path,'w'))

        json.dump(config, open(config_json,'w'))
        json.dump(hbuild, open(hbuild_json,'w'))
        json.dump(greedy, open(greedy_json,'w'))
        return params


    def build_atlas(self, configured_dir):

        output_dir=Path(configured_dir)
        
        def dmriatlas_proc(output_dir):
            with open(output_dir.joinpath('log.txt'),'w') as sys.stdout:
                import  dtiplayground.dmri.common as common   
                from dtiplayground.dmri.atlasbuilder import AtlasBuilder 
                logger = common.logger
                logger.setFilePointer(sys.stdout)
                config_path=output_dir.joinpath('common/config.yml')
                hbuild_path=output_dir.joinpath('common/h-build.yml')
                greedy_path=output_dir.joinpath('common/greedy.yml')

                bldr=AtlasBuilder(logger = logger)
                bldr.configure( output_dir=output_dir,
                                config_path=config_path,
                                hbuild_path=hbuild_path,
                                greedy_path=greedy_path)
           
                bldr.build()

        res = { 'execution_id' : utils.get_uuid(),
                'output_dir' : output_dir.__str__()
        }
        proc = Process(target= dmriatlas_proc,name=res['execution_id'], args=[output_dir])
        proc.start()
        res['pid']=proc.pid
        res['proc_name']=proc.name
        res['status']='running'
        json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
        proc.join()
        if proc.exitcode != 0 : 
            res['status']='failed'
            json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
            raise Exception("Error during running")
        else:
            res['status']='success'
            json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
        return res


    def getUITemplate(self):
        import dtiplayground.dmri.atlasbuilder.data as d
        template_fn = Path(d.__file__).parent.joinpath('template.json')
        template = json.load(open(template_fn,'r'))
        return template 



#!/usr/bin/env python
#
# SHBASECOPYRIGHT
#
# FSL installer script.
#
"""This is the FSL installation script, which can be used to install FSL.

This script must:

 - be able to be executed with Python 2.7 or newer.

 - be able to be executed in a "vanilla" Python environment, with no third
   party dependencies.

 - be self-contained, with no dependencies on any other modules (apart from
   the Python standard library).

 - be importable as a Python module - this script contains functions and
   classes that may be used by other scripts.
"""


from __future__ import print_function, division, unicode_literals

import functools      as ft
import os.path        as op
import subprocess     as sp
import textwrap       as tw
import                   argparse
import                   contextlib
import                   fnmatch
import                   getpass
import                   hashlib
import                   json
import                   logging
import                   os
import                   platform
import                   readline
import                   shlex
import                   shutil
import                   ssl
import                   sys
import                   tempfile
import                   threading
import                   time
import                   traceback

try:                import urllib.request as urlrequest
except ImportError: import urllib         as urlrequest


try:                import urllib.parse as urlparse
except ImportError: import                 urlparse

try:                import queue
except ImportError: import Queue as queue


PYVER = sys.version_info[:2]


log = logging.getLogger(__name__)


# this sometimes gets set to fslinstaller.pyc, so rstrip c
__absfile__ = op.abspath(__file__).rstrip('c')


__version__ = '3.3.0'
"""Installer script version number. This must be updated
whenever a new version of the installer script is released.
"""


DEFAULT_INSTALLATION_DIRECTORY = op.join(op.expanduser('~'), 'fsl')
"""Default FSL installation directory. """


FSL_RELEASE_MANIFEST = 'https://fsl.fmrib.ox.ac.uk/fsldownloads/' \
                       'fslconda/releases/manifest.json'
"""URL to download the FSL installer manifest file from. The installer
manifest file is a JSON file which contains information about available FSL
versions.

See the download_manifest function, and an example manifest file
in test/data/manifest.json, for more details.

A custom manifest URL can be specified with the -a/--manifest command-line
option.
"""


FSL_DEV_RELEASES = 'https://fsl.fmrib.ox.ac.uk/fsldownloads/' \
                   'fslconda/releases/devreleases.txt'
"""URL to the devreleases.txt file, which contains a list of available
internal/development FSL releases. See the download_dev_releases function
for more details.
"""


# List of modifiers which can be used to change how
# a message is printed by the printmsg function.
INFO      = 1
IMPORTANT = 2
QUESTION  = 3
PROMPT    = 4
WARNING   = 5
ERROR     = 6
EMPH      = 7
EMPHASIS  = 7
UNDERLINE = 8
RESET     = 9
ANSICODES = {
    INFO      : '\033[37m',         # Light grey
    IMPORTANT : '\033[92m',         # Green
    QUESTION  : '\033[36m\033[4m',  # Blue+underline
    PROMPT    : '\033[36m\033[1m',  # Bright blue+bold
    WARNING   : '\033[93m',         # Yellow
    ERROR     : '\033[91m',         # Red
    EMPHASIS  : '\033[1m',          # White+bold
    UNDERLINE : '\033[4m',          # Underline
    RESET     : '\033[0m',          # Used internally
}


def printmsg(*args, **kwargs):
    """Prints a sequence of strings formatted with ANSI codes. Expects
    positional arguments to be of the form::

        printable, ANSICODE, printable, ANSICODE, ...

    :arg log: Must be specified as a keyword argument. If True (default),
              the message is logged.

    All other keyword arguments are passed through to the print function.
    """

    args     = list(args)
    blockids = [i for i in range(len(args)) if (args[i] not in ANSICODES)]
    logmsg   = kwargs.pop('log', True)
    uncoded  = ''

    for i, idx in enumerate(blockids):
        if i == len(blockids) - 1:
            slc = slice(idx + 1, None)
        else:
            slc = slice(idx + 1, blockids[i + 1])

        msg      = args[idx]
        msgcodes = args[slc]
        msgcodes = [ANSICODES[c] for c in msgcodes]
        msgcodes = ''.join(msgcodes)
        uncoded += msg

        print('{}{}{}'.format(msgcodes, msg, ANSICODES[RESET]), end='')

    if len(blockids) > 0:
        print(**kwargs)
        if logmsg:
            log.debug(uncoded)

    sys.stdout.flush()


def prompt(promptmsg, *msgtypes, **kwargs):
    """Prompts the user for some input. msgtypes and kwargs are passed
    through to the printmsg function.
    """
    printmsg(promptmsg, *msgtypes, end='', log=False, **kwargs)

    if PYVER[0] == 2: response = raw_input(' ').strip()
    else:             response = input(    ' ').strip()

    log.debug('%s: %s', promptmsg, response)

    return response


def identify_platform():
    """Figures out what platform we are running on. Returns a platform
    identifier string - one of:

      - "linux-64" (Linux, x86_64)
      - "macos-64" (macOS, x86_64)
      - "macos-M1" (macOS, M1)

    Note that these identifiers are for FSL releases, and are not the
    same as the platform identifiers used by conda.
    """

    platforms = {
        ('linux',  'x86_64') : 'linux-64',
        ('darwin', 'x86_64') : 'macos-64',
        ('darwin', 'arm64')  : 'macos-M1',
    }

    system = platform.system().lower()
    cpu    = platform.machine()
    key    = (system, cpu)

    if key not in platforms:
        supported = ', '.join(['[{}, {}]' for s, c in platforms])
        raise Exception('This platform [{}, {}] is unrecognised or '
                        'unsupported! Supported platforms: {}'.format(
                            system, cpu, supported))

    return platforms[key]


def check_need_admin(dirname):
    """Returns True if dirname needs administrator privileges to write to,
    False otherwise.
    """
    # os.supports_effective_ids added in
    # python 3.3, so can't be used here
    return not os.access(dirname, os.W_OK | os.X_OK)


def get_admin_password(action='install FSL'):
    """Prompt the user for their administrator password. An Exception is raised
    if an incorrect password is entered three times.a

    :arg action: String which describes what the password is needed for, i.e.:
                 "Your administrator password is needed to {action}"
    :returns:    the validated administrator password
    """

    def validate_admin_password(password):
        proc = Process.sudo_popen(['true'], password, stdin=sp.PIPE)
        proc.communicate()
        return proc.returncode == 0

    msg = 'Your administrator password is needed to {}'.format(action)

    for attempt in range(3):
        if attempt == 0: msg = '{}:'.format(msg)
        else:            msg = '{} [attempt {} of 3]:'.format(msg, attempt + 1)
        printmsg(msg, IMPORTANT, end='')
        password = getpass.getpass('')
        valid    = validate_admin_password(password)

        if valid:
            printmsg('Password accepted', INFO)
            break
        else:
            printmsg('Incorrect password', WARNING)

    if not valid:
        raise Exception('Incorrect password')

    return password


def isstr(s):
    """Returns True if s is a string, False otherwise, Works on python 2.7
    and >=3.3.
    """
    try:              return isinstance(s, basestring)
    except Exception: return isinstance(s, str)


def match_any(s, patterns):
    """Test if the string s matches any of the fnmatch-style patterns.
    Returns the matched pattern, or None.
    """
    for pat in patterns:
        if fnmatch.fnmatch(s, pat):
            return pat
    return None


@contextlib.contextmanager
def tempdir(override_dir=None, change_into=True):
    """Returns a context manager which creates, changes into, and returns a
    temporary directory, and then deletes it on exit.

    If override_dir is not None, instead of creating and changing into a
    temporary directory, this function just changes into override_dir.
    """

    if override_dir is None: tmpdir = tempfile.mkdtemp()
    else:                    tmpdir = override_dir

    prevdir = os.getcwd()

    try:
        if change_into:
            os.chdir(tmpdir)
        yield tmpdir

    finally:
        if change_into:
            os.chdir(prevdir)
        if override_dir is None:
            shutil.rmtree(tmpdir)


@contextlib.contextmanager
def tempfilename(permissions=None, delete=True):
    """Returns a context manager which creates a temporary file, yields its
    name, then deletes the file on exit.
    """

    fname = None

    try:
        tmpf  = tempfile.NamedTemporaryFile(delete=False)
        fname = tmpf.name

        tmpf.close()

        if permissions:
            os.chmod(fname, permissions)

        yield fname

    finally:
        if delete and fname and op.exists(fname):
            os.remove(fname)


def sha256(filename, check_against=None, blocksize=1048576):
    """Calculate the SHA256 checksum of the given file. If check_against
    is provided, it is compared against the calculated checksum, and an
    error is raised if they are not the same.
    """

    hashobj = hashlib.sha256()

    with open(filename, 'rb') as f:
        while True:
            block = f.read(blocksize)
            if len(block) == 0:
                break
            hashobj.update(block)

    checksum = hashobj.hexdigest()

    if check_against is not None:
        if checksum != check_against:
            raise Exception('File {} does not match expected checksum '
                            '({})'.format(filename, check_against))

    return checksum


def clean_environ():
    """Return a dict containing a set of sanitised environment variables.

    All FSL and conda related variables are removed.
    """
    env = os.environ.copy()
    for v in list(env.keys()):
        if any(('FSL' in v, 'CONDA' in v, 'PYTHON' in v)):
            env.pop(v)
    return env


def install_environ(fsldir, username=None, password=None):
    """Returns a dict containing some environment variables that should
    be added to the shell environment when the FSL conda environment is
    being installed.
    """
    env = {}
    # post-link scripts call $FSLDIR/share/fsl/sbin/createFSLWrapper
    # (part of fsl/base), which will only do its thing if the following
    # env vars are set
    env['FSL_CREATE_WRAPPER_SCRIPTS'] = '1'
    env['FSLDIR']                     = fsldir

    # Make sure HTTP proxy variables, if set,
    # are available to the conda env command
    for v in ['http_proxy', 'https_proxy',
              'HTTP_PROXY', 'HTTPS_PROXY']:
        if v in os.environ:
            env[v] = os.environ[v]

    # FSL environments which source packages from the internal
    # FSL conda channel will refer to the channel as:
    #
    # http://${FSLCONDA_USERNAME}:${FSLCONDA_PASSWORD}/abc.com/
    #
    # so we need to set those variables
    if username: env['FSLCONDA_USERNAME'] = username
    if password: env['FSLCONDA_PASSWORD'] = password

    return env


def download_file(url,
                  destination,
                  progress=None,
                  blocksize=131072,
                  ssl_verify=True):
    """Download a file from url, saving it to destination. """

    def default_progress(downloaded, total):
        pass

    if progress is None:
        progress = default_progress

    log.debug('Downloading %s ...', url)

    # Path to local file
    if op.exists(url):
        url = 'file:' + urlrequest.pathname2url(op.abspath(url))

    # We create and use an unconfigured SSL
    # context to disable SSL verification.
    # Otherwise pass None causes urlopen to
    # use default behaviour. The context
    # argument is not available in py3.3
    kwargs = {}
    if (not ssl_verify) and (PYVER != (3, 3)):
        printmsg('Skipping SSL verification - this '
                 'is not recommended!', WARNING)
        kwargs['context'] = ssl.SSLContext(ssl.PROTOCOL_TLS)

    req = None

    try:
        # py2: urlopen result cannot be
        # used as a context manager
        req = urlrequest.urlopen(url, **kwargs)
        with open(destination, 'wb') as outf:

            try:             total = int(req.headers['content-length'])
            except KeyError: total = None

            downloaded = 0

            progress(downloaded, total)
            while True:
                block = req.read(blocksize)
                if len(block) == 0:
                    break
                downloaded += len(block)
                outf.write(block)
                progress(downloaded, total)

    finally:
        if req:
            req.close()


def download_manifest(url, workdir=None):
    """Downloads the installer manifest file, which contains information
    about available FSL versions, and the most recent version number of the
    installer (this script).

    The manifest file is a JSON file. Lines beginning
    with a double-forward-slash are ignored. See test/data/manifest.json
    for an example.

    This function modifies the manifest structure by adding a 'version'
    attribute to all FSL build entries.
    """

    log.debug('Downloading FSL installer manifest from %s', url)

    with tempdir(workdir):

        try:
            download_file(url, 'manifest.json')
        except Exception as e:
            log.debug('Error downloading FSL release manifest from %s',
                      url, exc_info=True)
            raise Exception('Unable to download FSL release manifest '
                            'from {} [{}]!'.format(url, str(e)))

        with open('manifest.json') as f:
            lines = f.readlines()

    # Drop comments
    lines = [l for l in lines if not l.lstrip().startswith('//')]

    manifest = json.loads('\n'.join(lines))

    # Add "version" to every build
    for version, builds in manifest['versions'].items():
        if version == 'latest':
            continue
        for build in builds:
            build['version'] = version

    return manifest


def download_dev_releases(url, workdir=None):
    """Downloads the FSL_DEV_RELEASES file. This file contains a list of
    available development manifest URLS. Returns a list of tuples, one
    for each development release, with each tuple containing:

      - URL to the manifest file
      - Most recent release tag
      - Date of the release
      - Commit hash (on the fsl/conda/manifest repository)
      - Branch name (on the fsl/conda/manifest repository)

    The list is sorted by date, newest first.
    """

    # parse a dev manifest file name, returning
    # a sequence containing the tage, date, commit
    # hash, and branch name. Dev manifest files
    # are named like so:
    #
    #   manifest-<tag>.<date>.<commit>.<branch>.json
    #
    # where <tag> is the tag of the most recent
    # public FSL release, and everything else is
    # self-explanatory.
    def parse_devrelease_name(url):
        name = urlparse.urlparse(url).path
        name = op.basename(name)
        name = name.lstrip('manifest-').rstrip('.json')
        # Awkward - the tag may have periods in it
        name = name.rsplit('.', 3)
        return name

    # list of (url, tag, date, commit, branch)
    devreleases = []

    with tempdir(workdir):

        try:
            download_file(url, 'devreleases.txt')
        except Exception as e:
            log.debug('Error downloading devreleases.txt from %s',
                      url, exc_info=True)
            raise Exception('Unable to download development manifest '
                            'list from {}!'.format(url))

        with open('devreleases.txt', 'rt') as f:
            urls = f.read().strip().split('\n')
            urls = [l.strip() for l in urls]

        for url in urls:
            devreleases.append([url] + parse_devrelease_name(url))

    # sort by date, newest first
    return sorted(devreleases, key=lambda r: r[2], reverse=True)


class Progress(object):
    """Simple progress reporter. Displays one of the following:

       - If both a value and total are provided, a progress bar is shown
       - If only a value is provided, a cumulative count is shown
       - If nothing is provided, a spinner is shown.

    Use as a context manager, and call the update method to report progress,
    e,g:

        with Progress('%') as p:
            for i in range(100):
                p.update(i + 1, 100)
    """

    def __init__(self,
                 label='',
                 transform=None,
                 fmt='{:.1f}',
                 total=None,
                 width=None):
        """Create a Progress reporter.

        :arg label:     Units (e.g. "MB", "%",)

        :arg transform: Function to transform values (see e.g.
                        Progress.bytes_to_mb)

        :arg fmt:       Template string used to format value / total.

        :arg total:     Maximum value - overrides the total value passed to
                        the update method.

        :arg width:     Maximum width, if a progress bar is displayed. Default
                        is to automatically infer the terminal width (see
                        Progress.get_terminal_width).
        """

        if transform is None:
            transform = Progress.default_transform

        self.width     = width
        self.fmt       = fmt.format
        self.total     = total
        self.label     = label
        self.transform = transform

        # used by the spin function
        self.__last_spin = None

    @staticmethod
    def default_transform(val, total):
        return val, total

    @staticmethod
    def bytes_to_mb(val, total):
        if val   is not None: val   = val   / 1048576
        if total is not None: total = total / 1048576
        return val, total

    @staticmethod
    def percent(val, total):
        if val is None or total is None:
            return val, total
        return 100 * (val / total), 100

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        printmsg('', log=False)

    def update(self, value=None, total=None):

        if total is None:
            total = self.total

        value, total = self.transform(value, total)

        if value is None and total is None:
            self.spin()
        elif value is not None and total is None:
            self.count(value)
        elif value is not None and total is not None:
            self.progress(value, total)

    def spin(self):

        symbols = ['|', '/', '-',  '\\']

        if self.__last_spin is not None: last = self.__last_spin
        else:                            last = symbols[-1]

        idx  = symbols.index(last)
        idx  = (idx + 1) % len(symbols)
        this = symbols[idx]

        printmsg(this, end='\r', log=False)
        self.__last_spin = this

    def count(self, value):

        value = self.fmt(value)

        if self.label is None: line = '{} ...'.format(value)
        else:                  line = '{}{} ...'.format(value, self.label)

        printmsg(line, end='\r', log=False)

    def progress(self, value, total):

        value = min(value, total)

        # arbitrary fallback of 50 columns if
        # terminal width cannot be determined
        if self.width is None: width = Progress.get_terminal_width(50)
        else:                  width = self.width

        fvalue = self.fmt(value)
        ftotal = self.fmt(total)
        suffix = '{} / {} {}'.format(fvalue, ftotal, self.label).rstrip()

        # +5: - square brackets around bar
        #     - space between bar and tally
        #     - space+spin at the end
        width     = width - (len(suffix) + 5)
        completed = int(round(width * (value  / total)))
        remaining = width - completed
        progress  = '[{}{}] {}'.format('#' * completed,
                                       ' ' * remaining,
                                       suffix)

        printmsg(progress, end='', log=False)
        printmsg(' ', end='', log=False)
        self.spin()
        printmsg(end='\r', log=False)


    @staticmethod
    def get_terminal_width(fallback=None):
        """Return the number of columns in the current terminal, or fallback
        if it cannot be determined.
        """
        # os.get_terminal_size added in python
        # 3.3, so we try it but fall back to
        # COLUMNS, or tput as a last resort.
        try:
            return shutil.get_terminal_size()[0]
        except Exception:
            pass

        try:
            return int(os.environ['COLUMNS'])
        except Exception:
            pass

        try:
            result = Process.check_output('tput cols', log_output=False)
            return int(result.strip())
        except Exception:
            return fallback


class Process(object):
    """Container for a subprocess.Popen object, allowing non-blocking
    line-based access to its standard output and error streams via separate
    queues, while logging all outputs.

    Don't create a Process directly - use one of the following static methods:
     - Process.check_output
     - Process.check_call
     - Process.monitor_progress
    """


    def __init__(self,
                 cmd,
                 admin=False,
                 password=None,
                 log_output=True,
                 print_output=False,
                 append_env=None,
                 **kwargs):
        """Run the specified command. Starts threads to capture stdout and
        stderr.

        :arg cmd:          Command to run - passed through shlex.split, then
                           passed to subprocess.Popen

        :arg admin:        Run the command with administrative privileges

        :arg password:     Administrator password - can be None if admin is
                           False.

        :arg log_output:   If True, the command and all of its stdout/stderr
                           are logged.

        :arg print_output: If True, the command and all of its stdout/stderr
                           are logged.

        :arg append_env:   Dictionary of additional environment to be set when
                           the command is run.

        :arg kwargs:       Passed to subprocess.Popen
        """

        self.cmd     = cmd
        self.stdoutq = queue.Queue()
        self.stderrq = queue.Queue()

        if log_output:
            log.debug('Running %s [as admin: %s]', cmd, admin)

        self.popen = Process.popen(cmd, admin, password,
                                   append_env=append_env, **kwargs)

        # threads for consuming stdout/stderr
        self.stdout_thread = threading.Thread(
            target=Process.forward_stream,
            args=(self.popen.stdout, self.stdoutq, cmd,
                  'stdout', log_output, print_output))
        self.stderr_thread = threading.Thread(
            target=Process.forward_stream,
            args=(self.popen.stderr, self.stderrq, cmd,
                  'stderr', log_output, print_output))

        self.stdout_thread.daemon = True
        self.stderr_thread.daemon = True
        self.stdout_thread.start()
        self.stderr_thread.start()


    def wait(self):
        """Waits for the process to terminate, then waits for the stdout
        and stderr consumer threads to finish.
        """
        self.popen.wait()
        self.stdout_thread.join()
        self.stderr_thread.join()


    @property
    def returncode(self):
        """Process return code. Returns None until the process has terminated,
        and the stdout/stderr consumer threads have finished.
        """
        if self.popen.returncode is None: return None
        if self.stdout_thread.is_alive(): return None
        if self.stderr_thread.is_alive(): return None
        return self.popen.returncode


    @staticmethod
    def check_output(cmd, *args, **kwargs):
        """Behaves like subprocess.check_output. Runs the given command, then
        waits until it finishes, and return its standard output. An error
        is raised if the process returns a non-zero exit code.

        :arg cmd: The command to run, as a string
        """

        proc = Process(cmd, *args, **kwargs)
        proc.wait()

        if proc.returncode != 0:
            raise RuntimeError('This command returned an error: ' + cmd)

        stdout = ''
        while True:
            try:
                stdout += proc.stdoutq.get_nowait()
            except queue.Empty:
                break

        return stdout


    @staticmethod
    def check_call(cmd, *args, **kwargs):
        """Behaves like subprocess.check_call. Runs the given command, then
        waits until it finishes. An error is raised if the process returns a
        non-zero exit code.

        :arg cmd: The command to run, as a string
        """
        proc = Process(cmd, *args, **kwargs)
        proc.wait()
        if proc.returncode != 0:
            raise RuntimeError('This command returned an error: ' + cmd)


    @staticmethod
    def monitor_progress(cmd, total=None, *args, **kwargs):
        """Runs the given command(s), and shows a progress bar under the
        assumption that cmd will produce "total" number of lines of output.

        :arg cmd:      The commmand to run as a string, or a sequence of
                       multiple commands.

        :arg total:    Total number of lines of standard output to expect.

        :arg timeout:  Refresh rate in seconds. Must be passed as a keyword
                       argument.

        :arg progfunc: Function which returns a number indicating how far
                       the process has progressed.  If provided, this
                       function is called, instead of standard output
                       lines being monitored. The function is passed a
                       reference to the Process object. Must be passed as a
                       keyword argument.
        """

        timeout  = kwargs.pop('timeout',  0.5)
        progfunc = kwargs.pop('progfunc', None)

        if total is None: label = None
        else:             label = '%'

        if progfunc is None:
            nlines = [0]
            def progfunc(proc):
                try:
                    _         = proc.stdoutq.get_nowait()
                    nlines[0] = nlines[0] + 1
                except queue.Empty:
                    pass
                return nlines[0]

        if isstr(cmd): cmds = [cmd]
        else:          cmds =  cmd

        with Progress(label=label,
                      fmt='{:.0f}',
                      transform=Progress.percent) as prog:

            progcount = 0 if total else None

            for cmd in cmds:

                proc = Process(cmd, *args, **kwargs)
                prog.update(progcount, total)

                while proc.returncode is None:
                    time.sleep(timeout)
                    progcount = progfunc(proc) if total else None
                    prog.update(progcount, total)
                    proc.popen.poll()

                # force progress bar to 100% when finished
                if proc.returncode == 0:
                    prog.update(total, total)
                else:
                    raise RuntimeError('This command returned '
                                       'an error: ' + cmd)


    @staticmethod
    def forward_stream(stream,
                       queue,
                       cmd,
                       streamname,
                       log_output,
                       print_output):
        """Reads lines from stream and pushes them onto queue until popen
        is finished. Logs every line.

        :arg stream:       stream to forward
        :arg queue:        queue.Queue to push lines onto
        :arg cmd:          string - the command that is running
        :arg streamname:   string - 'stdout' or 'stderr'
        :arg log_output:   If True, log all stdout/stderr.
        :arg print_output: If True, print all stdout/stderr.
        """

        while True:
            line = stream.readline().decode('utf-8')
            if line == '':
                break
            queue.put(line)
            if log_output:
                log.debug(' [%s]: %s', streamname, line.rstrip())
            if print_output:
                print(' [{}]: {}'.format(streamname, line.rstrip()))


    @staticmethod
    def popen(cmd, admin=False, password=None, append_env=None, **kwargs):
        """Runs the given command via subprocess.Popen, as administrator if
        requested.

        :arg cmd:        The command to run, as a string

        :arg admin:      Whether to run with administrative privileges

        :arg pssword:    Administrator password. Only required if admin is
                         True.

        :arg append_env: Dictionary of additional environment to be set when
                         the command is run.

        :arg kwargs:     Passed to subprocess.Popen. stdin, stdout, and stderr
                         will be silently clobbered

        :returns:        The subprocess.Popen object.
        """

        admin = admin and os.getuid() != 0

        cmd              = shlex.split(cmd)
        kwargs['stdin']  = sp.PIPE
        kwargs['stdout'] = sp.PIPE
        kwargs['stderr'] = sp.PIPE

        if admin:
            proc = Process.sudo_popen(cmd, password, append_env, **kwargs)
        else:
            # if append_env has been specified,
            # add it to the normal env option.
            if append_env is not None:
                env = kwargs.get('env', os.environ.copy())
                env.update(append_env)
                kwargs['env'] = env

            proc = sp.Popen(cmd, **kwargs)

        return proc


    @staticmethod
    def sudo_popen(cmd, password, append_env=None, **kwargs):
        """Runs "sudo cmd" using subprocess.Popen. Used by Process.popen.
        Assumes that kwargs contains stdin=sp.PIPE
        """

        # sudo will not necessarily propagate environment
        # variables, and there is no guarantee that the
        # sudo -E option will work. So here we create a
        # wrapper shell script with "export VAR=VALUE"
        # statements for all environment variables that
        # are set.
        if append_env is None:
            append_env = {}

        # Make the wrapper script delete itself
        # after the command has been executed.
        with tempfilename(0o755, delete=False) as wrapper:
            with open(wrapper, 'wt') as f:
                f.write('#!/usr/bin/env sh\n')
                f.write('set -e\n')
                f.write('thisfile=$0\n')
                f.write('thisdir=$(cd $(dirname $0) && pwd)\n')
                for k, v in append_env.items():
                    f.write('export {}="{}"\n'.format(k, v))
                # shlex.join not available in py27
                f.write(' '.join(cmd) + '\n')
                f.write('cd ${thisdir} && rm ${thisfile}\n')

        cmd  = ['sudo', '-S', '-k', wrapper]
        proc = sp.Popen(cmd, **kwargs)
        proc.stdin.write('{}\n'.format(password).encode())
        proc.stdin.flush()
        return proc


@ft.total_ordering
class Version(object):
    """Class to represent and compare version strings.  Accepted version
    strings are of the form W.X.Y.Z, where W, X, Y, and Z are all integers.
    """
    def __init__(self, verstr):
        # Version identifiers for official FSL
        # releases will have up to four
        # components (X.Y.Z.W), but we accept
        # any number of (integer) components,
        # as internal releases may have more.
        components = []

        # ignore a leading "v", e.g. v1.2.3
        verstr = verstr.lower()
        if verstr.startswith('v'):
            verstr = verstr[1:]

        for comp in verstr.split('.'):
            try:              components.append(int(comp))
            except Exception: break

        self.components = components
        self.verstr     = verstr

    def __str__(self):
        return self.verstr

    def __eq__(self, other):
        for sn, on in zip(self.components, other.components):
            if sn != on:
                return False
        return len(self.components) == len(other.components)

    def __lt__(self, other):
        for p1, p2 in zip(self.components, other.components):
            if p1 < p2: return True
            if p1 > p2: return False
        return len(self.components) < len(other.components)


class Context(object):
    """Bag of information and settings created in the main function, and passed
    around this script.

    Several settings are lazily evaluated on first access, but once evaluated,
    their values are immutable.
    """

    def __init__(self, args, destdir=None, action='install FSL'):
        """Create the context with the argparse.Namespace object containing
        parsed command-line arguments.

        :arg args:    argparse.Namespace containing command-line arguments
        :arg destdir: Installation directory. If not provided, read from
                      args.dest, or read from the user,
        :arg action:  Passed to get_admin_password as a prompt.
        """

        self.args  = args
        self.shell = op.basename(os.environ.get('SHELL', 'sh')).lower()

        # These attributes are updated on-demand via
        # the property accessors defined below, or are
        # all updated via the finalise_settings method.
        self.__platform         = None
        self.__manifest         = None
        self.__devmanifest      = None
        self.__candidate_builds = None
        self.__build            = None
        self.__destdir          = destdir
        self.__need_admin       = None
        self.__admin_password   = None
        self.__action           = action

        # If the destination directory already exists,
        # and the user chooses to overwrite it, it is
        # moved so that, if the installation fails, it
        # can be restored. The new path is stored
        # here - refer to overwrite_destdir.
        self.old_destdir = None

        # The download_fsl_environment function stores
        # the path to the FSL conda environment file
        # and list of conda channels
        self.environment_file     = None
        self.environment_channels = None

        # The config_logging function stores the path
        # to the fslinstaller log file here.
        self.logfile = None


    def finalise_settings(self):
        """Finalise values for all information and settings in the Context.
        """
        self.manifest
        self.candidate_builds
        self.platform
        self.build
        self.destdir
        self.need_admin
        self.admin_password


    @property
    def platform(self):
        """The platform we are running on, e.g. "linux-64", "macos-64",
        "macos-M1". This identifier is used to determine which FSL build to
        install.

        Note that this function may report a different platform identifier than
        the actual platform - specifically, if running on a M1 mac, and there
        is no M1 FSL build for the requested FSL version, this function will
        report "macos-64". This is because some older FSL releases do not have
        M1 builds available.
        """
        if self.__platform is None:
            plat = identify_platform()

            # if M1, check that we have a suitable
            # FSL build, falling back to x86 if not.
            if plat == 'macos-M1':
                candidates = self.candidate_builds
                if not any(c['platform'] == 'macos-M1' for c in candidates):
                    plat = 'macos-64'

            self.__platform = plat

        return self.__platform


    @property
    def candidate_builds(self):
        """Query the manifest and return a list of available builds for the
        requested FSL release, for all platforms.
        """
        if self.__candidate_builds is not None:
            return self.__candidate_builds

        # defaults to "latest" if
        # not specified by the user
        fslversion = self.args.fslversion
        if fslversion is None:
            fslversion = 'latest'

        if fslversion not in self.manifest['versions']:
            available = ', '.join(self.manifest['versions'].keys())
            raise Exception(
                'FSL version "{}" is not available - available '
                'versions: {}'.format(fslversion, available))

        if fslversion == 'latest':
            fslversion = self.manifest['versions']['latest']

        self.__candidate_builds = list(self.manifest['versions'][fslversion])

        return self.__candidate_builds


    @property
    def build(self):
        """Returns a suitable FSL build (a dictionary entry from the FSL
        installer manifest) for the target platform.

        The returned dictionary has the following elements:
          - 'version'      FSL version.
          - 'platform':    Platform identifier (e.g. 'linux-64')
          - 'environment': Environment file to download
          - 'sha256':      Checksum of environment file
          - 'output':      Number of lines of expected output, for reporting
                           progress
        """

        if self.__build is not None:
            return self.__build

        # Find refs to a suitable build for this
        # platform. We assume that there is only
        # one default build for each platform.
        # in the list of builds for a given FSL
        # version.
        candidates = self.candidate_builds
        build      = None

        for candidate in candidates:
            if candidate['platform'] == self.platform:
                build = candidate
                break
        else:
            raise Exception(
                'Cannot find a version of FSL matching '
                'platform {}'.format(self.platform))

        printmsg('FSL {} selected for installation'.format(build['version']))

        self.__build = build
        return build


    @property
    def destdir(self):
        """Installation directory. If not specified at the command line, the
        user is prompted to enter a directory.
        """

        if self.__destdir is not None:
            return self.__destdir

        # The loop below validates the destination directory
        # both when specified at commmand line or
        # interactively.  In either case, if invalid, the
        # user is re-prompted to enter a new destination.
        destdir = None
        if self.args.dest is not None: response = self.args.dest
        else:                          response = None

        while destdir is None:

            if response is None:
                printmsg('\nWhere do you want to install FSL?',
                         IMPORTANT, EMPHASIS)
                printmsg('Press enter to install to the default location '
                         '[{}]\n'.format(DEFAULT_INSTALLATION_DIRECTORY), INFO)
                response = prompt('FSL installation directory [{}]:'.format(
                    DEFAULT_INSTALLATION_DIRECTORY), QUESTION, EMPHASIS)
                response = response.rstrip(op.sep)

                if response == '':
                    response = DEFAULT_INSTALLATION_DIRECTORY

            response  = op.expanduser(op.expandvars(response))
            response  = op.abspath(response)
            parentdir = op.dirname(response)
            if op.exists(parentdir):
                destdir = response
            else:
                printmsg('Destination directory {} does not '
                         'exist!'.format(parentdir), ERROR)
                response = None

        self.__destdir = destdir
        return self.__destdir


    @property
    def conda(self):
        """Return a path to the ``conda`` or ``mamba`` executable. """
        # If mamba is present, prefer it over conda
        candidates = [
            op.join(self.destdir, 'bin', 'mamba'),
            op.join(self.destdir, 'bin', 'conda')]
        for c in candidates:
            if op.exists(c):
                return c


    @property
    def need_admin(self):
        """Returns True if administrator privileges will be needed to install
        FSL.
        """
        if self.__need_admin is not None:
            return self.__need_admin
        parentdir = op.dirname(self.destdir)
        self.__need_admin = check_need_admin(parentdir)
        return self.__need_admin


    @property
    def admin_password(self):
        """Returns the user's administrator password, prompting them if needed.
        """
        if self.__admin_password is not None:
            return self.__admin_password
        # need_admin may be None or False,
        # so don't rely on truthiness.
        if self.__need_admin == False:
            return None
        self.__admin_password = get_admin_password(self.__action)


    @property
    def manifest(self):
        """Returns the FSL installer manifest as a dictionary. """

        if self.__manifest is None:
            if self.devmanifest is not None:
                self.args.manifest = self.devmanifest

            self.__manifest = download_manifest(self.args.manifest,
                                                self.args.workdir)
        return self.__manifest


    @property
    def devmanifest(self):
        """Returns a URL to a development manifest to use for installation.
        This will only return a value if the --devrelease or --devlatest
        options are active.

        If this is the case, the FSL_DEV_RELEASES file is downloaded - this
        file contains a list of available development manifest URLS. The
        user is then prompted to choose which development manifest to use
        for the installation, unless --devlatest is active, in which case
        the newest manifest is selected.
        """
        if not self.args.devrelease:
            return None
        if self.__devmanifest == 'na':
            return None
        elif self.__devmanifest is not None:
            return self.__devmanifest

        devreleases = download_dev_releases(FSL_DEV_RELEASES,
                                            self.args.workdir)

        if len(devreleases) == 0:
            self.__devmanifest = 'na'
            return None

        self.__devmanifest = prompt_dev_release(devreleases,
                                                self.args.devlatest)

        return self.__devmanifest


    def run(self, process_func, *args, **kwargs):
        """Run a command via a static Process method.  Handles sudo/
        administrator authentication, and ensures that the shell
        environment in which the command is executed is sanitised.

        Can be used with Process.check_call, Process.check_output, and
        Process.monitor_progress. For example:

            ctx = Context(...)
            ctx.run(Process.check_call, 'my_command')
            ctx.run(Process.monitor_progress, 'my_command', total=100)
        """

        process_func = ft.partial(process_func, *args, **kwargs)

        # Clear any environment variables that refer to
        # existing FSL or conda installations, and ensure
        # that some specific FSL environment variables
        # are set while the command is running.  See
        # clean_environ and install_environ for more
        # details, and see Process.sudo_popen regarding
        # append_env.
        env        = clean_environ()
        append_env = install_environ(self.destdir,
                                     self.args.username,
                                     self.args.password)
        return process_func(admin=self.need_admin,
                            password=self.admin_password,
                            env=env,
                            append_env=append_env)


def check_rosetta_status(ctx):
    """Called from the main routine, before installation is attempted.  If
    running on a M1 macos machine, and a x86 version of FSL has been selected
    for installation, checks whether rosetta emulation is enabled. If so,
    does nothing further. Otherwise, prints a message and exits.
    """

    if not all((identify_platform() == 'macos-M1',
                ctx.platform        == 'macos-64')):
        return

    # Using the strategy discussed at
    # https://forum.latenightsw.com/t/possible-for-a-script-\
    #   to-test-whether-rosetta-2-is-installed/3207/6
    #
    # The pkgutil command should return 0 if
    # rosetta is installed, non-0 otherwise.
    try:
        Process.check_output('pkgutil --files com.apple.pkg.RosettaUpdateAuto')
    except RuntimeError:
        printmsg('Rosetta emulation does not appear to be enabled!\n', ERROR)
        printmsg('Enable rosetta emulation, and then run this installer '
                 'again. You can enable rosetta emulation by running this '
                 'command:\n', INFO)
        printmsg('  /usr/sbin/softwareupdate --install-rosetta '
                 '--agree-to-license\n', IMPORTANT)
        printmsg('Aborting installation', ERROR)
        sys.exit(1)
    # pkgutil command not found - should
    # never happen, but print a warning
    # just in case
    except Exception as e:
        printmsg('An error occurred calling the pkgutil command - this '
                 'may not be a problem, so I\'ll attempt to proceed '
                 'with the installation. ({}'.format(e), WARNING)


def list_available_versions(manifest):
    """Lists available FSL versions. """
    printmsg('Available FSL versions:', EMPHASIS)
    for version in manifest['versions']:
        if version == 'latest':
            continue
        printmsg(version, IMPORTANT, EMPHASIS)
        for build in manifest['versions'][version]:
            printmsg('  {}'.format(build['platform']), EMPHASIS, end=' ')
            printmsg(build['environment'], INFO)


def prompt_dev_release(devreleases, latest):
    """Prompts the user to select a development release.

    :arg devreleases: List of development releases, as returned by
                      download_dev_releases.
    :arg latest:      If True, the latest develeopment release is returned.
    """

    if len(devreleases) == 0:
        return None

    # automatically choose latest dev manifest?
    if latest:
        return devreleases[0][0]

    # show the user a list, ask them which one they want
    printmsg('Available development releases:', EMPHASIS)
    for i, (url, tag, date, commit, branch) in enumerate(devreleases):
        printmsg('  [{}]: {} [{} commit {}]'.format(
            i + 1, date, branch, commit), IMPORTANT)
    while True:
        selection = prompt('Which release would you like to '
                           'install? [1]:', PROMPT)
        if selection == '':
            selection = '1'
        try:
            selection = int(selection) - 1
        except Exception:
            continue
        if selection >= 0 and selection < len(devreleases):
            break
    return devreleases[selection][0]


def download_fsl_environment(ctx):
    """Downloads the environment specification file for the selected FSL
    version.

    Internal/development FSL versions may source packages from the internal
    FSL conda channel, which requires a username+password to authenticate.

    These are referred to in the environment file as ${FSLCONDA_USERNAME}
    and ${FSLCONDA_PASSWORD}.

    If the user has not provided a username+password on the command-line, they
    are prompted for them.

    The downloaded environment file may be modified - if the (hidden)
    --exclude_package option has been used.
    """

    build        = ctx.build
    url          = build['environment']
    checksum     = build.get('sha256', None)

    printmsg('Downloading FSL environment specification '
             'from {}...'.format(url))
    fname = url.split('/')[-1]
    download_file(url, fname)
    ctx.environment_file = op.abspath(fname)
    if (checksum is not None) and (not ctx.args.no_checksum):
        sha256(fname, checksum)

    # Environment files for internal/dev FSL versions
    # will list the internal FSL conda channel with
    # ${FSLCONDA_USERNAME} and ${FSLCONDA_PASSWORD}
    # as placeholders for the username/password.
    with open(fname, 'rt') as f:
        need_auth = '${FSLCONDA_USERNAME}' in f.read()

    # We need a username/password to access the internal
    # FSL conda channel. Prompt the user if they haven't
    # provided credentials.
    if need_auth and (ctx.args.username is None):
        printmsg('A username and password are required to install '
                 'this version of FSL.', WARNING, EMPHASIS)
        ctx.args.username = prompt('Username:').strip()
        ctx.args.password = getpass.getpass('Password: ').strip()

    # Conda expands environment variables within a
    # .condarc file, but *not* within an environment.yml
    # file. So to authenticate to our internal channel
    # without storing credentials anywhere in plain text,
    # we *move* the channel list from the environment.yml
    # file into $FSLDIR/.condarc.
    #
    # Here we extract the channels from the environment
    # file, and save them to ctx.environment_channels.
    # The install_miniconda function will then add the
    # channels to $FSLDIR/.condarc.
    #
    # We also remove any packages that the user has
    # requested to exclude from the installation.
    copy     = '.' + op.basename(ctx.environment_file)
    channels = []

    shutil.move(ctx.environment_file, copy)
    with open(copy,                 'rt') as inf, \
         open(ctx.environment_file, 'wt') as outf:

        in_channels_section = False

        for line in inf:

            # start of channels list
            if line.strip() == 'channels:':
                in_channels_section = True
                continue

            if in_channels_section:
                # end of channels list
                if not line.strip().startswith('-'):
                    in_channels_section = False
                else:
                    channels.append(line.split()[-1])
                    continue

            # Include/exclude packages upon user request
            pkgname = line.strip(' -').split()[0]
            exclude = match_any(pkgname, ctx.args.exclude_package)
            if exclude:
                log.debug('Excluding package %s (matched '
                          '--exclude_package %s)', line, exclude)
            else:
                outf.write(line)

    ctx.environment_channels = channels


def download_miniconda(ctx):
    """Downloads the miniconda/miniforge installer and saves it as
    "miniconda.sh".

    This function assumes that it is run within a temporary/scratch directory.
    """

    if ctx.args.miniconda is None:
        metadata = ctx.manifest['miniconda'][ctx.platform]
        url      = metadata['url']
        checksum = metadata['sha256']
    else:
        url      = ctx.args.miniconda
        checksum = None

    # Download
    printmsg('Downloading miniconda from {}...'.format(url))
    with Progress('MB', transform=Progress.bytes_to_mb) as prog:
        download_file(url, 'miniconda.sh', prog.update)
    if (not ctx.args.no_checksum) and (checksum is not None):
        sha256('miniconda.sh', checksum)


def install_miniconda(ctx):
    """Downloads the miniconda/miniforge installer, and installs it to the
    destination directory.

    This function assumes that it is run within a temporary/scratch directory.
    """

    metadata = ctx.manifest['miniconda'][ctx.platform]
    output   = metadata.get('output', '').strip()

    if output == '': output = None
    else:            output = int(output)

    # Install
    printmsg('Installing miniconda at {}...'.format(ctx.destdir))
    cmd = 'bash miniconda.sh -b -p {}'.format(ctx.destdir)
    ctx.run(Process.monitor_progress, cmd, total=output)

    # Avoid WSL filesystem issue
    # https://github.com/conda/conda/issues/9948
    cmd = 'find {} -type f -exec touch {{}} +'.format(ctx.destdir)
    ctx.run(Process.check_call, cmd)

    # Generate $FSLDIR/.condarc which contains
    # some default/fixed conda settings
    condarc = generate_condarc(ctx.environment_channels,
                               ctx.args.skip_ssl_verify)
    with open('.condarc', 'wt') as f:
        f.write(condarc)

    ctx.run(Process.check_call, 'cp -f .condarc {}'.format(ctx.destdir))


def generate_condarc(channels, skip_ssl_verify=False):
    """Called by install_miniconda. Generates content for a .condarc file to
    be saved in $FSLDIR/.condarc. This file contains some default values, and
    also enforces some settings so that they cannot be overridden by the
    user. For example. the list of conda channels is configured so that it
    cannot be overridden by a user's ~/.condarc file.

    See the following web pages for more details:
     - https://docs.conda.io/projects/conda/en/latest/user-guide/configuration/use-condarc.html
     - https://www.anaconda.com/blog/conda-configuration-engine-power-users
     - https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-channels.html
    """

    # Create .condarc config file
    condarc = tw.dedent("""
    # FSL conda configuration file, auto-generated by the fslinstaller script.
    #
    # WARNING: This file may be automatically re-generated
    # without warning, so it is recommended that any custom
    # conda settings are stored elsewhere. Refer to the conda
    # documentation for more guidance:
    #
    # https://conda.io/projects/conda/en/latest/user-guide/configuration/use-condarc.html
    # https://www.anaconda.com/blog/conda-configuration-engine-power-users
    # https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-channels.html

    # Try and make package downloads more robust
    remote_read_timeout_secs:    240
    remote_connect_timeout_secs: 20
    remote_max_retries:          10
    remote_backoff_factor:       5
    safety_checks:               warn

    # Disable caching of remote channel repodata.
    # This is a hack which is combined with the
    # WSL1 filesystem hack in the fslinstaller
    # script - because we have modified file
    # timestamps, conda will assume that its
    # channel repodata cache is up to date, and
    # will not bother refreshing it in the
    # commands that we run in the install_fsl
    # function. When we remove the WSL1 hack from
    # the fslinstaller, we can remove this config
    # setting.
    # https://github.com/conda/conda/issues/9948
    local_repodata_ttl: 0

    # Channel priority is important. In older versions
    # of FSL we placed the FSL conda channel at the
    # bottom (lowest priority) for legacy reasons (to
    # ensure that conda-forge versions of e.g. VTK were
    # preferred over legacy FSL conda versions).
    #
    # Use final/top/bottom marks to prevent the channel
    # priority order being modified by user ~/.condarc
    # configuration files.
    channel_priority: strict #!final
    """)

    if skip_ssl_verify:
        printmsg('Configuring conda to skip SSL verification '
                 '- this is not recommended!', WARNING)
        condarc += tw.dedent("""
        # Disable SSL verification when accessing
        # conda channels over https:// (the
        # fslinstaller script was called with
        # --skip_ssl_verify). NOT RECOMMENDED.
        ssl_verify: false
        """)

    channels = list(channels)
    if len(channels) > 0:
        channels[0]  += ' #!top'
        channels[-1] += ' #!bottom'
    condarc      += '\nchannels: #!final\n'
    for channel in channels:
        condarc += ' - {}\n'.format(channel)

    return condarc


def get_install_fsl_progress_reporting_method(ctx):
    """Figure out which reporting mechansim to use for reporting progress
    whilst FSL is being installed. The mechanism that is used has changed
    a few times.

    Returns a tuple containing values to pass to the Progress.monitor_progress
    function, either of which may be None if progress cannot be reported:

      - an integer value to pass as the total
      - a function to pass as the progfunc.
    """

    # We calculate installation progress in
    # one of a few ways, as we have changed
    # the mechanism a few times.  The
    # 'output/install' field in the manifest
    # gives us information about how to
    # report installation progress.
    fslver     = ctx.build['version']
    progparams = ctx.build.get('output', {}).get('install', None)
    pkgdir     = op.join(ctx.destdir, 'pkgs')

    # The first method (version 1) involves
    # progress reporting by monitoring number of
    # lines of standard output produced by
    # "conda env install". This is set to None,
    # as it is the default behaviour of the
    # Progress.monitor_progress function.
    progress_v1 = None

    # The second method involves progress
    # reporting by monitoring the number of
    # package files created in $FSLDIR/pkgs/
    def progress_v2(_):
        pkgs = os.listdir(pkgdir)
        pkgs = [p for p in pkgs if p.endswith('.conda') or p.endswith('.bz2')]
        return len(pkgs)

    progresses      = {}
    progresses['1'] = None
    progresses['2'] = progress_v2

    progval  = None
    progfunc = None

    # The output field may be either a
    # string, in which case we assume
    # version 2, or a dictionary containing
    # the progress reporting version, and
    # an integer value.
    if isstr(progparams):
        progval  = int(progparams)
        progfunc = progresses['2']

    # output field is a dict - versioned
    # progress reporting
    elif isinstance(progparams, dict):
        progval  = int(progparams['value'])
        progfunc = progresses[progparams['version']]

    return progval, progfunc


def install_fsl(ctx):
    """Install FSL into ctx.destdir (which is assumed to be a miniconda
    installation.

    This function assumes that it is run within a temporary/scratch directory.
    """

    progval, progfunc = get_install_fsl_progress_reporting_method(ctx)

    # We install FSL simply by running conda env
    # update -f env.yml.
    cmd = ctx.conda + ' env update -n base -f ' + ctx.environment_file
    printmsg('Installing FSL into {}...'.format(ctx.destdir))
    ctx.run(Process.monitor_progress, cmd,
            timeout=1, total=progval, progfunc=progfunc)


def finalise_installation(ctx):
    """Performs some finalisation tasks. Includes:
      - Saving the installed version to $FSLDIR/etc/fslversion
      - Saving this installer script and the environment file to
        $FSLDIR/etc/
    """
    with open('fslversion', 'wt') as f:
        f.write(ctx.build['version'])

    etcdir = op.join(ctx.destdir, 'etc')
    cmds   = [
        'cp fslversion {}'.format(etcdir),
        'cp {} {}'        .format(ctx.environment_file, etcdir)]

    for cmd in cmds:
        ctx.run(Process.check_call, cmd)


def post_install_cleanup(ctx):
    """Cleans up the FSL directory after installation. """

    cmd   = ctx.conda + ' clean -y --all'
    ctx.run(Process.check_call, cmd)


def patch_file(filename, searchline, numlines, content):
    """Used by configure_shell and configure_matlab. Adds to, modifies,
    or creates the specified file.

    If a line matching searchline is found in the file, numlines (starting
    from searchline) are replaced with content.

    Otherwise, content is appended to the end of the file.
    """

    content = content.split('\n')

    if op.isfile(filename):
        with open(filename) as f:
            lines = [l.strip() for l in f.readlines()]
    else:
        lines = []

    # replace block
    try:
        idx   = lines.index(searchline)
        lines = lines[:idx] + content + lines[idx + numlines:]

    # append to end
    except ValueError:
        lines = lines + [''] + content

    with open(filename, 'wt') as f:
        f.write('\n'.join(lines))


def configure_shell(shell, homedir, fsldir):
    """Configures the user's shell environment (e.g. ~/.bash_profile).

    :arg shell:   User's shell (taken from the $SHELL environment variable
    :arg homedir: User's home directory, presumed to contain shell profile
                  file(s).
    :arg fsldir:  FSL installation directory
    """

    bourne_shells  = ['sh', 'bash', 'zsh', 'dash']
    csh_shells     = ['csh', 'tcsh']

    # we edit the first file that exists in
    # the list of candidate profile files.
    # They are attached as an attribute of
    # this function just for testing purposes
    # (see after function definition)
    shell_profiles = configure_shell.shell_profiles

    # DO NOT CHANGE the format of these configurations -
    # they are kept exactly as-is for compatibility with
    # legacy FSL installations, i.e. so we can modify
    # profiles with an existing configuration from older
    # FSL versions
    bourne_cfg = tw.dedent("""
    # FSL Setup
    FSLDIR={fsldir}
    PATH=${{FSLDIR}}/share/fsl/bin:${{PATH}}
    export FSLDIR PATH
    . ${{FSLDIR}}/etc/fslconf/fsl.sh
    """).format(fsldir=fsldir).strip()

    csh_cfg = tw.dedent("""
    # FSL Setup
    setenv FSLDIR {fsldir}
    setenv PATH ${{FSLDIR}}/share/fsl/bin:${{PATH}}
    source ${{FSLDIR}}/etc/fslconf/fsl.csh
    """).format(fsldir=fsldir).strip()

    if shell not in bourne_shells + csh_shells:
        printmsg('Shell {} not recognised - skipping environment '
                 'setup'.format(shell), WARNING, EMPHASIS)
        return

    if shell in bourne_shells: cfg = bourne_cfg
    else:                      cfg = csh_cfg

    # find the profile file to edit
    profile    = None
    candidates = [op.join(homedir, p)
                  for p in shell_profiles[shell]]
    for candidate in candidates:
        if op.isfile(candidate):
            profile = candidate
            break

    # if no candidate profile files
    # exist, fall back to the first one
    if profile is None:
        profile = candidates[0]

    printmsg('Adding FSL configuration to {}'.format(profile))

    patch_file(profile, '# FSL Setup', len(cfg.split('\n')), cfg)
configure_shell.shell_profiles = {'sh'   : ['.profile'],
                                  'bash' : ['.bash_profile', '.profile'],
                                  'dash' : ['.bash_profile', '.profile'],
                                  'zsh'  : ['.zprofile'],
                                  'csh'  : ['.cshrc'],
                                  'tcsh' : ['.tcshrc']}


def configure_matlab(homedir, fsldir):
    """Creates/appends FSL configuration code to ~/Documents/MATLAB/startup.m.
    """

    # DO NOT CHANGE the format of this configuration -
    # see in-line comments in configure_shell.
    cfg = tw.dedent("""
    % FSL Setup
    setenv( 'FSLDIR', '{fsldir}' );
    setenv('FSLOUTPUTTYPE', 'NIFTI_GZ');
    fsldir = getenv('FSLDIR');
    fsldirmpath = sprintf('%s/etc/matlab',fsldir);
    path(path, fsldirmpath);
    clear fsldir fsldirmpath;
    """).format(fsldir=fsldir).strip()

    matlab_dir = op.expanduser(op.join(homedir, 'Documents', 'MATLAB'))
    startup_m  = op.join(matlab_dir, 'startup.m')

    if not op.exists(matlab_dir):
        os.makedirs(matlab_dir)

    printmsg('Adding FSL configuration to {}'.format(startup_m))

    patch_file(startup_m, '% FSL Setup', len(cfg.split('\n')), cfg)


def self_update(manifest, workdir, checksum):
    """Checks to see if a newer version of the installer (this script) is
    available and if so, downloads it to a temporary file, and runs it in
    place of this script.
    """

    thisver   = Version(__version__)
    latestver = Version(manifest['installer']['version'])

    if latestver <= thisver:
        log.debug('Installer is up to date (this version: %s, '
                  'latest version: %s)', thisver, latestver)
        return

    log.debug('New version of installer is available '
              '(%s) - self-updating', latestver)

    tmpf = tempfile.NamedTemporaryFile(
        prefix='new_fslinstaller', delete=False, dir=workdir)
    tmpf.close()
    tmpf = tmpf.name

    download_file(manifest['installer']['url'], tmpf)

    if checksum:
        try:
            sha256(tmpf, manifest['installer']['sha256'])
        except Exception as e:
            printmsg('New installer file does not match expected '
                     'checksum! Skipping update.', WARNING)
            return

    # Don't try and update again - if for some
    # reason the online manifest reports a newer
    # version than what is available, we would
    # otherwise enter into an infinite loop.
    cmd = [sys.executable, tmpf] + sys.argv[1:] + ['--no_self_update']
    log.debug('Running new installer: %s', cmd)
    os.execv(sys.executable, cmd)


def overwrite_destdir(ctx):
    """Called by main if the destination directory already exists. Asks the
    user if they want to overwrite it. If they do, or if the --overwrite
    option was specified, the directory is moved, and then deleted after
    the installation succeeds.

    This function assumes that it is run within a temporary/scratch directory.
    """

    if not ctx.args.overwrite:
        printmsg()
        printmsg('Destination directory [{}] already exists!'
                 .format(ctx.destdir), WARNING, EMPHASIS)
        response = prompt('Do you want to overwrite it [y/N]?',
                          QUESTION, EMPHASIS)
        if response.lower() not in ('y', 'yes'):
            printmsg('Aborting installation', ERROR, EMPHASIS)
            sys.exit(1)

    # generate a unique name for the old
    # destination directory (to avoid
    # collisions if using the same workdir
    # repeatedly)
    i = 0
    while True:
        ctx.old_destdir = op.abspath('old_destdir{}'.format(i))
        i              += 1
        if not op.exists(ctx.old_destdir):
            break

    printmsg('Deleting directory {}'.format(ctx.destdir), IMPORTANT)
    ctx.run(Process.check_call,
            'mv {} {}'.format(ctx.destdir, ctx.old_destdir))


def parse_args(argv=None, include=None):
    """Parse command-line arguments, returns an argparse.Namespace object.

    :arg argv:    Command-line arguments.

    :arg include: List of arguments to parse. May be used by other scripts
                  which re-use some of the routines defined in this script.
                  The resulting argparse.Namespace object will contain values
                  of None for all arguments that are not included.
    """

    username = os.environ.get('FSLCONDA_USERNAME', None)
    password = os.environ.get('FSLCONDA_PASSWORD', None)

    options = {
        # regular options
        'version'      : ('-v', {'action'  : 'version',
                                 'version' : __version__}),
        'dest'         : ('-d', {'metavar' : 'DESTDIR'}),
        'overwrite'    : ('-o', {'action'  : 'store_true'}),
        'listversions' : ('-l', {'action'  : 'store_true'}),
        'no_env'       : ('-n', {'action'  : 'store_true'}),
        'no_shell'     : ('-s', {'action'  : 'store_true'}),
        'no_matlab'    : ('-m', {'action'  : 'store_true'}),
        'fslversion'   : ('-V', {'default' : 'latest'}),

        # hidden options
        'username'        : (None, {'default' : username}),
        'password'        : (None, {'default' : password}),
        'no_checksum'     : (None, {'action'  : 'store_true'}),
        'skip_ssl_verify' : (None, {'action'  : 'store_true'}),
        'workdir'         : (None, {}),
        'homedir'         : (None, {'default' : op.expanduser('~')}),
        'devrelease'      : (None, {'action'  : 'store_true'}),
        'devlatest'       : (None, {'action'  : 'store_true'}),
        'manifest'        : (None, {}),
        'miniconda'       : (None, {}),
        'no_self_update'  : (None, {'action'  : 'store_true'}),
        'exclude_package' : (None, {'action'  : 'append'}),
    }

    if include is None:
        include = list(options.keys())

    helps = {
        'version'      : 'Print installer version number and exit',
        'listversions' : 'List available FSL versions and exit',
        'dest'         : 'Install FSL into this folder (default: '
                         '{})'.format(DEFAULT_INSTALLATION_DIRECTORY),
        'overwrite'    : 'Delete existing destination directory if it exists, '
                         'without asking',
        'no_env'       : 'Do not modify your shell or MATLAB configuration '
                         'implies --no_shell and --no_matlab)',
        'no_shell'     : 'Do not modify your shell configuration',
        'no_matlab'    : 'Do not modify your MATLAB configuration',
        'fslversion'   : 'Install this specific version of FSL',

        # Username / password for accessing
        # internal FSL conda channel, if an
        # internal/development release is being
        # installed. If not set, will be read from
        # FSLCONDA_USERNAME/FSLCONDA_PASSWORD
        # environment variables.
        'username'        : argparse.SUPPRESS,
        'password'        : argparse.SUPPRESS,

        # Do not automatically update the installer script,
        'no_self_update'  : argparse.SUPPRESS,

        # Install a development release. This
        # option will cause the installer to
        # download the devrelreases.txt file,
        # which contains a list of available
        # internal/development manifests. The
        # user will be prompted to choose one,
        # which will be propagated on to the
        # --manifest option. If --devlatest
        # is used, the most recent developmet
        # release is automatically selected.
        'devrelease'      : argparse.SUPPRESS,
        'devlatest'       : argparse.SUPPRESS,

        # Path/URL to alternative FSL release
        # manifest.
        'manifest'        : argparse.SUPPRESS,

        # Install miniconda from this path/URL,
        # instead of the one specified in the
        # FSL release manifest
        'miniconda'       : argparse.SUPPRESS,

        # Print debugging messages
        'debug'           : argparse.SUPPRESS,

        # Disable SHA256 checksum validation of downloaded files
        'no_checksum'     : argparse.SUPPRESS,

        # Store temp files in this directory
        # rather than in a temporary directory
        'workdir'         : argparse.SUPPRESS,

        # Treat this directory as user's home
        # directory, for the purposes of shell
        # configuration. Must already exist.
        'homedir'         : argparse.SUPPRESS,

        # Configure conda to skip SSL verification.
        # Not recommended.
        'skip_ssl_verify' : argparse.SUPPRESS,

        # Do not install packages matching this
        # fnmatch-style wildcard pattern. Can
        # be used multiple times.
        'exclude_package' : argparse.SUPPRESS,
    }

    # parse args
    parser = argparse.ArgumentParser()
    for option in include:
        shortflag, kwargs = options[option]
        flags             = ['--{}'.format(option)]
        if shortflag is not None:
            flags.insert(0, shortflag)
        parser.add_argument(*flags, help=helps[option], **kwargs)

    # parse_known_args so that newly added
    # args are ignored by older versions,
    # but will be parsed after self_update
    args = parser.parse_known_args(argv)[0]

    # add placeholder values for excluded args
    for option in options.keys():
        if option not in include:
            setattr(args, option, None)

    # alternate home dir (for debugging)
    if args.homedir is not None:
        args.homedir = op.abspath(args.homedir)
        if not op.isdir(args.homedir):
            printmsg('Home directory {} does not exist!'.format(args.homedir),
                     ERROR, EMPHASIS)
            sys.exit(1)

    # don't modify shell profile
    if args.no_env:
        args.no_shell  = True
        args.no_matlab = True

    # use workdir rather than a tempdir
    if args.workdir is not None:
        args.workdir = op.abspath(args.workdir)
        if not op.exists(args.workdir):
            os.mkdir(args.workdir)

    # manifest takes priority over devrelease/devlatest
    if args.manifest is not None:
        args.devrelease = False
        args.devlatest  = False

    if args.manifest is None:
        args.manifest = FSL_RELEASE_MANIFEST

    if args.devlatest:
        args.devrelease = True

    if args.exclude_package is None:
        args.exclude_package = []

    # accept local path for manifest and environment
    if args.manifest is not None and op.exists(args.manifest):
        args.manifest = op.abspath(args.manifest)

    # accept local path for miniconda installer
    if args.miniconda is not None and op.exists(args.miniconda):
        args.miniconda = op.abspath(args.miniconda)

    return args


def config_logging(prefix='fslinstaller_', logdir=None):
    """Configures logging. Log messages are directed to
    $TMPDIR/fslinstaller_<unique_token>.log, or
    logdir/fslinstaller_<unique_token>.log
    """
    if logdir is None:
        logdir = tempfile.gettempdir()

    # Use a unique name for the log file
    # (important for multi-user systems)
    logfilef, logfile = tempfile.mkstemp(prefix=prefix,
                                         suffix='.log',
                                         dir=logdir)
    os.close(logfilef)

    handler   = logging.FileHandler(logfile)
    formatter = logging.Formatter(
        '%(asctime)s %(filename)s:%(lineno)4d: %(message)s', '%H:%M:%S')
    handler.setFormatter(formatter)
    log.addHandler(handler)
    log.setLevel(logging.DEBUG)

    return logfile


@contextlib.contextmanager
def handle_error(ctx):
    """Used by main as a context manager around the main installation steps.
    If an error occurs, prints some messages, performs some clean-up/
    restoration tasks, and exits.
    """

    try:
        yield

    except Exception as e:
        printmsg('\nERROR occurred during installation!', ERROR, EMPHASIS)
        printmsg('    {}\n'.format(e), INFO)

        # send traceback to log file
        tb = traceback.format_tb(sys.exc_info()[2])
        log.debug(''.join(tb))

        if op.exists(ctx.destdir):
            printmsg('Removing failed installation directory '
                     '{}'.format(ctx.destdir), WARNING)
            ctx.run(Process.check_call, 'rm -r ' + ctx.destdir)

        # overwrite_destdir moves the existing
        # destdir to a temp location, so we can
        # restore it if the installation fails
        if not op.exists(ctx.destdir) and (ctx.old_destdir is not None):
            printmsg('Restoring contents of {}'.format(ctx.destdir),
                     WARNING)
            ctx.run(Process.check_call,
                    'mv {} {}'.format(ctx.old_destdir, ctx.destdir))

        printmsg('\nFSL installation failed!', ERROR, EMPHASIS)
        printmsg('The log file may contain some more information to help '
                 'you diagnose the problem: {}'.format(ctx.logfile), ERROR)
        sys.exit(1)


def main(argv=None):
    """Installer entry point. Downloads and installs miniconda and FSL, and
    configures the user's environment.
    """

    printmsg('FSL installer version:', EMPHASIS, UNDERLINE, end='')
    printmsg(' {}'.format(__version__))
    printmsg('Press CTRL+C at any time to cancel installation', INFO)

    if os.getuid() == 0:
        printmsg('Running the installer script as root user is discouraged! '
                 'You should run this script as a regular user - you will be '
                 'asked for your administrator password if required.',
                 WARNING, EMPHASIS)

    args    = parse_args(argv)
    logfile = config_logging(logdir=args.workdir)

    log.debug(' '.join(sys.argv))
    printmsg('Installation log file: {}\n'.format(logfile), INFO)

    ctx         = Context(args)
    ctx.logfile = logfile

    if not args.no_self_update:
        self_update(ctx.manifest, args.workdir, not args.no_checksum)

    if args.listversions:
        list_available_versions(ctx.manifest)
        sys.exit(0)

    try:
        ctx.finalise_settings()
    except Exception as e:
        printmsg('An error has occurred: {}'.format(e), ERROR)
        sys.exit(1)

    check_rosetta_status(ctx)

    with tempdir(args.workdir):

        # Ask the user if they want to overwrite
        # an existing installation
        if op.exists(ctx.destdir):
            overwrite_destdir(ctx)

        download_fsl_environment(ctx)

        printmsg('\nInstalling FSL in {}\n'.format(ctx.destdir), EMPHASIS)
        with handle_error(ctx):
            download_miniconda(ctx)
            install_miniconda(ctx)
            install_fsl(ctx)
            finalise_installation(ctx)
            post_install_cleanup(ctx)

    if not args.no_shell:
        configure_shell(ctx.shell, args.homedir, ctx.destdir)
    if not args.no_matlab:
        configure_matlab(args.homedir, ctx.destdir)

    printmsg('\nFSL successfully installed\n', IMPORTANT)
    if not args.no_shell:
        printmsg('Open a new terminal, or log out and log back in, '
                 'for the environment changes to take effect.', INFO)


if __name__ == '__main__':
    sys.exit(main())



from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class DTITractStat(ExternalToolWrapper):
    def __init__(self,binary_path= None, **kwargs):
        super().__init__(binary_path, **kwargs)
        self.binary_path=None
        if binary_path is not None:
            self.binary_path=binary_path
        elif 'softwares' in kwargs:
            self.binary_path=kwargs['softwares']['dtitractstat']['path']

    """
    DTI Fiber Tract Statistics Tool Parameters:

    - input_fiber_file: --input_fiber_file, -i
      Input fiber file in .vtk format

    - output_stats_file: --ouput_stats_file, -o
      Output statistics file in .fvp format along the fiber

    - output_parametrized_fiber_file: --output_parametrized_fiber_file, -f
      Output parametrized fiber file

    - output_original_fibers_parametrized: --output_original_fibers_parametrized, -d
      Output original fibers with arclength to plane

    - plane_file: --plane_file
      File specifying the plane

    - auto_plane_origin: --auto_plane_origin
      Way to calculate auto plane origin (default: cog)

    - window_file: --window_file
      CSV file with histogram values for the chosen kernel window

    - image_space: --image_space
      Set if the fiber file is in image space (default: False)

    - step_size: --step_size
      Step size for sampling the fiber file (default: 1)

    - processing_rodent: --rodent
      Set when processing rodent images (default: False)

    - window: --window
      Window value for visualization (-1 means no window chosen, default: -1)

    - bandwidth: --bandwidth
      Bandwidth or std for the kernel used in noise model (default: 1)

    - parameter_list: --parameter_list
      List of scalar diffusion properties (default: fa,md,ad,l2,l3,fro,rd,ga)

    - scalarName: --scalarName
      Optional scalar properties present in the fiber file

    - noise_model: --noise_model
      Noise model used (default: gaussian)

    - stat_type: --stat_type
      Type of maximum likelihood estimate (default: mean)

    - q_perc: --quantile_percentage
      Quantile percentage [0,100] (default: 50)

    - useNonCrossingFibers: --use_non_crossing_fibers
      Use non-crossing fibers (default: False)

    - removeNanFibers: --remove_nan_fibers
      Remove fibers that contain NaN values (default: False)

    - removeCleanFibers: --remove_clean_fiber
      Remove clean fiber files saved on the hard drive (default: False)
    """

    def run(self, input_fiber_file: str, output_stats_file: str, options=[]):
        arguments=[]
        arguments+=['--input_fiber_file',input_fiber_file]
        arguments+=['-o',output_stats_file]
        arguments+=options
        self.setArguments(arguments)
        return self.execute(arguments)



#!python

import sys
import os 
import traceback
import shutil
import socket
from pathlib import Path
import argparse
from argparse import RawTextHelpFormatter
import yaml
sys.path.append(Path(__file__).resolve().parent.parent.__str__())  ## this line is for development
import dtiplayground
import dtiplayground.dmri.common as common
from dtiplayground.config import INFO as info
from dtiplayground.api.server import DTIPlaygroundServer
from dtiplayground.dmri.playground.app import DMRIPlaygroundApp

logger=common.logger.write 
color= common.Color

## utils

def next_free_port(port=6543, max_port=6999 ):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while port <= max_port:
        try:
            sock.bind(('', port))
            sock.close()
            return port
        except OSError:
            port += 1
    raise IOError('no free ports')

### unit functions

def command_install_tools(args):
    ## reparametrization
    options = {
        'config_dir' : args.config_dir,
        'output_dir' : args.output_dir,
        'clean_install': args.clean_install,
        'no_fsl' : args.no_fsl,
        'install_only' : args.install_only,
        'build' : args.build,
        'no_remove' : args.no_remove,
    }
    app = DMRIPlaygroundApp(options['config_dir'])
    app.install_tools(options)
    return True

def command_init(args):
    ## reparametrization
    options = {
        'config_dir' : args.config_dir,
        'log' : args.log,
        'execution_id' : args.execution_id,
        'no_verbosity' : args.no_verbosity,
        'no_log_timestamp' : args.no_log_timestamp,
        'version' : args.version,
        'tools_dir' : args.tools_dir,
    }
    app = DMRIPlaygroundApp(options['config_dir'])
    return app.initialize(options)
    

def command_server(args):
    config = {
        "config_dir": args.config_dir,
        "host" : args.host,
        "port" : next_free_port(int(args.port)),
        "static_page_dir" : Path(args.directory).resolve().__str__(),
        "browser" : args.browser,
        "debug" : args.debug
    }

    if config['browser']: 
        import webbrowser
        from threading import Timer
        def open_browser():
            webbrowser.open('http://{}:{}'.format(config['host'],config['port']))
        Timer(1, open_browser).start();
        
    app = DMRIPlaygroundApp(config['config_dir'])
    app.run(config)
    
    return True

### Arguments 

def get_args():
    version=info['dmriplayground']['version']
    logger("VERSION : {}".format(str(version)))
    config_dir=Path.home().joinpath('.niral-dti').resolve()
    # ## read template
    uid, ts = common.get_uuid(), common.get_timestamp()

    ### Argument parsers

    parser=argparse.ArgumentParser(prog="dpg",
                                   formatter_class=RawTextHelpFormatter,
                                   description="The dtiplayground is an integrated DWI processing platform",
                                   epilog="Written by SK Park (sangkyoon_park@med.unc.edu) , Johanna Dubos (johannadubos32@gmail.com) , Neuro Image Research and Analysis Laboratories, University of North Carolina @ Chapel Hill , United States, 2021")
    subparsers=parser.add_subparsers(help="Commands")
    
    ## init command
    parser_init=subparsers.add_parser('init',help='Initialize configurations')
    parser_init.set_defaults(func=command_init)

    ## software install command
    parser_install_tools=subparsers.add_parser('install-tools',help='Install DTIPlaygroundTools')
    parser_install_tools.add_argument('-o','--output-dir', help="output directory", default="$HOME/.niral-dti")
    parser_install_tools.add_argument('-c','--clean-install', help="Remove existing files", default=False, action="store_true")
    parser_install_tools.add_argument('-b','--build', help="Build DTIPlaygroundTools", default=False, action="store_true")
    parser_install_tools.add_argument('--no-remove', help="Do not remove source and build files after installation", default=False,action="store_true")
    parser_install_tools.add_argument('--no-fsl', help="Do not install FSL", default=False, action="store_true")
    parser_install_tools.add_argument('--install-only', help="Do not update current software paths", default=False, action="store_true")
    parser_install_tools.set_defaults(func=command_install_tools)

    ## DPG Server
    parser_server=subparsers.add_parser('serve',help='DTIPlayground Server')
    parser_server.add_argument('--host', help="Host", default="127.0.0.1")
    parser_server.add_argument('-p','--port', help="Port", default=6543)
    parser_server.add_argument('-d','--directory', help="Static Page Path", default=str(config_dir.joinpath('static/spa')))
    parser_server.add_argument('--browser', help="Launch browser at start up", default=False, action="store_true")
    parser_server.add_argument('--debug', help="Debug mode", default=False, action="store_true")
    parser_server.set_defaults(func=command_server)

    ## log related
    parser.add_argument('--config-dir',help='Configuration directory',default=str(config_dir))
    parser.add_argument('--log',help='log file',default=str(config_dir.joinpath('log.txt')))
    parser.add_argument('--execution-id',help='execution id',default=uid,type=str)
    parser.add_argument('--no-log-timestamp',help='Remove timestamp in the log', default=False, action="store_true")
    parser.add_argument('--no-verbosity',help='Do not show any logs in the terminal', default=False, action="store_true")
    parser.add_argument('-v','--version', help="Show version", default=False,action="store_true")
    parser.add_argument('--tools-dir', help="Initialize with specific tool directory", default=None)

    ## if no parameter is furnished, exit with printing help
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args=parser.parse_args()
    if args.version:
        sys.exit(1)

    return args 

## threading environment
args=get_args()
if hasattr(args,'num_threads'):
    os.environ['OMP_NUM_THREADS']=str(args.num_threads) ## this should go before loading any dipy function. 
    os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = str(args.num_threads) ## for ANTS threading

if __name__=='__main__':
    try:
        common.logger.setTimestamp(True)
        result=args.func(args)
        exit(0)
    except Exception as e:
        common.logger.setVerbosity(True)
        msg=traceback.format_exc()
        logger(msg,color.ERROR)
        exit(-1)
    finally:
        pass





#!python

import os
import sys
import argparse
from argparse import RawTextHelpFormatter
import traceback
from pathlib import Path 
import yaml 
sys.path.append(Path(__file__).parent.parent.__str__())
### for dev
from argparse import RawTextHelpFormatter
import dtiplayground.dmri.atlasbuilder.utils as utils 
from dtiplayground.dmri.atlasbuilder import AtlasBuilder 
import dtiplayground.dmri.common as common

from dtiplayground.config import INFO as info
from dtiplayground.dmri.atlasbuilder.app import DMRIAtlasBuilderApp

logger=common.logger.write 
color= common.Color


def command_init(args):
    ## reparametrization
    options = {
        'config_dir' : args.config_dir,
        'log' : args.log,
        'execution_id' : args.execution_id,
        'no_verbosity' : args.no_verbosity,
        'no_log_timestamp' : args.no_log_timestamp,
        'version' : args.version,
        'tools_dir' : args.tools_dir,
    }
    app = DMRIAtlasBuilderApp(options['config_dir'])
    return app.initialize(options)

def build_atlas(args):
    params_dir = args.params_dir

    if params_dir is None:
        options={
            "output_dir" : args.output_dir,
            "config_dir" : args.config_dir,
            "hbuild_path" : args.hbuild,
            "config_path" : args.config,
            "greedy_path" : args.greedy
        }
        app = DMRIAtlasBuilderApp(options['config_dir'])
        app.run(options)        
    else:
        build_dir = Path(params_dir)
        output_dir = args.output_dir
        config_dir = args.config_dir
        hbuild_path = build_dir.joinpath('common/h-build.json').__str__()
        config_path = build_dir.joinpath('common/config.json').__str__()
        greedy_path = build_dir.joinpath('common/greedy.json').__str__()
        options={
            "output_dir" : output_dir,
            "config_dir" : config_dir,
            "hbuild_path" : hbuild_path,
            "config_path" : config_path,
            "greedy_path" : greedy_path
        }
        app = DMRIAtlasBuilderApp(options['config_dir'])
        app.run(options)        


def build_atlas_dir(args):
    build_dir = Path(args.build_dir)
    output_dir = build_dir.__str__()
    config_dir = args.config_dir
    hbuild_path = build_dir.joinpath('common/h-build.json').__str__()
    config_path = build_dir.joinpath('common/config.json').__str__()
    greedy_path = build_dir.joinpath('common/greedy.json').__str__()
    options={
        "output_dir" : output_dir,
        "config_dir" : config_dir,
        "hbuild_path" : hbuild_path,
        "config_path" : config_path,
        "greedy_path" : greedy_path
    }
    app = DMRIAtlasBuilderApp(options['config_dir'])
    app.run(options)


def get_args():
    version = info['dmriatlas']['version']
    logger("VERSION : {}".format(str(version)))
    config_dir=Path.home().joinpath('.niral-dti').resolve()
    # ## read template
    module_help_str=None
    if config_dir.exists() and config_dir.joinpath('config.yml').exists() and config_dir.joinpath('environment.yml').exists():
        config,environment = load_configurations(str(config_dir))
        template_path=config_dir.joinpath(config['protocol_template_path'])
        template=yaml.safe_load(open(template_path,'r'))
        available_modules=template['options']['execution']['pipeline']['candidates']
        available_modules_list=["{}".format(x['value'])  for x in available_modules if x['description']!="Not implemented"]
        module_help_str="Avaliable Modules := \n" + " , ".join(available_modules_list)
    uid, ts = common.get_uuid(), common.get_timestamp()

    ### Argument parsers

    parser=argparse.ArgumentParser(prog="dmriatlas",
                                   formatter_class=RawTextHelpFormatter,
                                   description="dmriatlas is a tool to make DTI Atlas.",
                                   epilog="Written by SK Park (sangkyoon_park@med.unc.edu)  ,Neuro Image Research and Analysis Laboratories, University of North Carolina @ Chapel Hill , United States, 2021")
    subparsers=parser.add_subparsers(help="Commands")

    ## init command
    parser_init=subparsers.add_parser('init',help='Initialize configurations')
    parser_init.set_defaults(func=command_init)
    
    ## build command
    parser_build=subparsers.add_parser('build',help='Build Atlas')
    parser_build.add_argument('-o','--output-dir', help='Output directory',required=True,type=str)
    parser_build_param_dir= parser_build.add_argument_group('from-dir')
    parser_build_param_dir.add_argument('-p','--params-dir', help='Parameter directory', default=None, type=str)
    parser_build_param = parser_build.add_argument_group('from-files')
    parser_build_param.add_argument('-c','--config',help='configuration file',default='config.yml',type=str)
    parser_build_param.add_argument('-b','--hbuild',help='hierarchical build file', default='h-build.yml', type=str)
    parser_build_param.add_argument('-g','--greedy',help='Greedy atlas parameter xml file', default="greedy.xml", type=str)
    parser_build.set_defaults(func=build_atlas)

    ## build dir command
    parser_build_dir=subparsers.add_parser('build-dir',help='Build Atlas with parameterized directory')
    parser_build_dir.add_argument('build_dir', help='Build directory having parameters',type=str)
    parser_build_dir.set_defaults(func=build_atlas_dir)

    parser.add_argument('--config-dir',help='Configuration directory',default=str(config_dir))
    parser.add_argument('--log',help='log file',default=str(config_dir.joinpath('log.txt')))
    parser.add_argument('--execution-id',help='execution id',default=uid,type=str)
    parser.add_argument('--no-log-timestamp',help='Remove timestamp in the log', default=False, action="store_true")
    parser.add_argument('--no-verbosity',help='Do not show any logs in the terminal', default=False, action="store_true")
    parser.add_argument('-v','--version', help="Show version", default=False,action="store_true")
    parser.add_argument('--tools-dir', help="Initialize with specific tool directory", default=None)
    
    ## if no parameter is furnished, exit with printing help
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args=parser.parse_args()
    if args.version:
        sys.exit(1)

    return args 

## threading environment
args=get_args()
if hasattr(args,'num_threads'):
    os.environ['OMP_NUM_THREADS']=str(args.num_threads) ## this should go before loading any dipy function. 
    os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = str(args.num_threads) ## for ANTS threading

if __name__=='__main__':
    try:
        common.logger.setTimestamp(True)
        result=args.func(args)
        exit(0)
    except Exception as e:
        common.logger.setVerbosity(True)
        msg=traceback.format_exc()
        logger(msg,color.ERROR)
        exit(-1)
    finally:
        pass




from dtiplayground.dmri.common.tools.base import ExternalToolWrapper


class DTIProcess(ExternalToolWrapper):
    def __init__(self, binary_path=None, **kwargs):
        super().__init__(binary_path, **kwargs)
        self.binary_path = None
        if binary_path is not None:
            self.binary_path = binary_path
        elif 'softwares' in kwargs:
            self.binary_path = kwargs['softwares']['dtiprocess']['path']

    def measure_scalars(self, inputfile, outputfile, scalar_type='FA', options=[]):
        assert (scalar_type.lower() in ['fa', 'md', 'ad', 'rd'])
        scalar_opt_key = scalar_type.lower()
        scalar_opt_map = {'fa': '-f', 'md': '-m', 'ad': '--lambda1_output', 'rd': '--RD_output'} # maps the scalar_type to the command line option
        scalar_opt = scalar_opt_map[scalar_opt_key]
        arguments = ['--dti_image', inputfile, scalar_opt, outputfile] + options
        self.setArguments(arguments)
        return self.execute(arguments)

    def measure_scalar_list(self, inputfile, outputFileStem, scalar_list=['FA', 'MD', 'AD', 'RD'], options=[]):
        arguments = ['--dti_image', inputfile]
        for scalar in scalar_list:
            assert (scalar.lower() in ['fa', 'md', 'ad', 'rd'])
            scalar_opt_key = scalar.lower()
            scalar_opt_map = {'fa': '-f', 'md': '-m', 'ad': '--lambda1_output', 'rd': '--RD_output'}
            scalar_opt = scalar_opt_map[scalar_opt_key]
            arguments += [scalar_opt, outputFileStem + '_' + scalar + '.nrrd']
        arguments += options
        self.setArguments(arguments)
        return self.execute(arguments)